
from __future__ import annotations
from typing import Optional, Union, List
from functools import reduce
import operator
import numpy
from ._C import Size, SizeValue, dtype
from ..naming import Naming


_dtype = dtype


class Tensor:
  def __init__(self, *args, dtype: Optional[Union[str, _dtype]] = None, numpy_array: Optional[numpy.ndarray] = None):
    if dtype is not None and isinstance(dtype, _dtype):
      dtype = dtype.name
    if args and isinstance(args[0], Tensor):
      assert len(args) == 1
      assert numpy_array is None
      numpy_array = args[0]._numpy_buffer.copy()
      shape = args[0].shape
    elif args and isinstance(args[0], (tuple, list)):
      assert len(args) == 1
      shape = tuple(args[0])
    else:
      shape = args
    assert isinstance(shape, tuple) and all(isinstance(dim, int) for dim in shape)
    shape = tuple([d if isinstance(d, SizeValue) else SizeValue(d) for d in shape])
    if numpy_array is not None:
      if dtype is not None:
        numpy_array = numpy_array.astype(dtype)
      else:
        dtype = str(numpy_array.dtype)
      assert numpy_array.shape == shape
    if dtype is None:
      dtype = "float32"
    self._shape = shape
    self._numpy_buffer = numpy.zeros(shape, dtype=dtype) if numpy_array is None else numpy_array
    self.dtype = _dtype(dtype)
    Naming.get_instance().register_tensor(self)

  def __repr__(self):
    return f"<{self.__class__.__name__} {self.returnn_naming_entry.repr_content()}>"

  @property
  def returnn_naming_entry(self):
    return Naming.get_instance().register_tensor(self)

  def dim(self):
    return len(self._shape)

  @property
  def ndim(self):
    return self.dim()

  def size(self, dim: Optional[int] = None) -> Size[SizeValue, ...]:
    if dim is not None:
      return self._shape[dim]
    return Size(self._shape)

  @property
  def shape(self) -> Size[SizeValue, ...]:
    return self.size()

  @property
  def data(self):
    return self  # TODO?

  def numel(self) -> int:
    """
    :returns: the total number of elements in the :attr:`input` tensor.
    """
    return reduce(operator.mul, self.shape, 1)

  def type(self, dtype=None, non_blocking=False, **kwargs) -> Union[_dtype, Tensor]:
    if dtype is None:
      return self.dtype
    from .nn.functional import cast
    return cast(self, dtype=dtype)

  def type_as(self, tensor) -> Tensor:
    return self.type(tensor.type())

  def to(self, opt):
    return self  # ignore

  def contiguous(self):
    return self  # ignore

  def flatten(self):
    from .nn.functional import flatten
    return flatten(self)

  def view(self, *shape):
    from .nn.functional import reshape
    return reshape(self, shape)

  def unsqueeze(self, dim: int):
    assert -len(self._shape) <= dim <= len(self._shape)
    if dim < 0:
      dim += len(self._shape) + 1
    assert 0 <= dim <= len(self._shape)
    return self.view(*(self._shape[:dim] + (1,) + self._shape[dim:]))

  def transpose(self, dim0: int, dim1: int):
    from .nn.functional import transpose
    return transpose(self, dim0=dim0, dim1=dim1)

  def expand(self, *sizes):
    from .nn.functional import expand
    return expand(self, sizes)

  def resize_(self, *sizes, memory_format=None):
    # memory format is ignored
    assert len(self.shape) == 0  # not implemented otherwise
    self._shape = sizes
    self._numpy_buffer = numpy.zeros(sizes, dtype=self.dtype.canonical_name)

  def new_zeros(self, *size, dtype=None, device=None, requires_grad=False):
    if len(size) == 1 and isinstance(size[0], (list, tuple)):
      size = size[0]
    if dtype is None:
      dtype = self.dtype
    else:
      dtype = _dtype(dtype)
    from .nn.functional import zeros
    return zeros(size, dtype=dtype)

  def new_empty(self, *size, dtype=None, device=None, requires_grad=False):
    # use new_zeros here to avoid errors by uninitialized memory
    return self.new_zeros(size, dtype, device, requires_grad)

  def new(self, *args, dtype=None, device=None, requires_grad=False):
    if args and isinstance(args[0], Tensor):
      assert len(args) == 1
      assert args[0].type() == self.type()
      return args[0]
    else:
      shape = args
      return self.new_empty(shape, dtype, device, requires_grad)

  def copy_(self, source: Tensor):
    self._numpy_buffer = source.view(*self._shape).type(self.dtype)._numpy_buffer.copy()

  def normal_(self, mean=0, std=1):
    from .nn.init import normal_
    normal_(self, mean=mean, std=std)
    return self

  def uniform_(self, a=0, b=1):
    from .nn.init import uniform_
    uniform_(self, a=a, b=b)
    return self

  def zero_(self):
    self.fill_(0)
    return self

  def fill_(self, x):
    if not self._shape:  # scalar
      self._numpy_buffer = numpy.array(x, dtype=self.dtype.name)
    else:
      self._numpy_buffer[:] = x

  def detach(self):
    return self  # TODO use stop_gradient?

  def numpy(self):
    return self._numpy_buffer

  def float(self):
    from .nn.functional import cast
    return cast(self, "float32")

  def int(self):
    from .nn.functional import cast
    return cast(self, "int32")

  def abs(self):
    from .nn.functional import abs
    return abs(self)

  def log(self):
    from .nn.functional import log
    return log(self)

  def sigmoid(self):
    from .nn.functional import sigmoid
    return sigmoid(self)

  def chunk(self, chunks: int, dim: int = 0) -> List[Tensor]:
    from .nn.functional import chunk
    return chunk(self, chunks=chunks, dim=dim)

  def __getitem__(self, item):
    assert self._shape  # cannot subscript a scalar
    if isinstance(item, int):
      from .nn import Gather
      return Gather(dim=0, pos=item)(self)
    elif isinstance(item, slice):
      from .nn import Slice
      return Slice(axis=0, start=item.start, stop=item.stop, step=item.step)(self)
    elif isinstance(item, tuple):
      assert len(item) == self.ndim
      from .nn import Slice
      out = self
      for ax, ax_slice in enumerate(item):
        assert isinstance(ax_slice, slice)
        if not ax_slice.start and not ax_slice.stop and ax_slice.step in {1, None}:
          continue
        out = Slice(axis=ax, start=ax_slice.start, stop=ax_slice.stop, step=ax_slice.step)(out)
      return out
    else:
      raise NotImplementedError

  def __setitem__(self, key, value):
    self._numpy_buffer[key] = value

  def __add__(self, other):
    from .nn.functional import add
    return add(self, other)

  def __sub__(self, other):
    from .nn.functional import sub
    return sub(self, other)

  def __mul__(self, other):
    from .nn.functional import mul
    return mul(self, other)

  def __truediv__(self, other):
    from .nn.functional import truediv
    return truediv(self, other)

  def __radd__(self, other):
    from .nn.functional import add
    return add(other, self)

  def __rsub__(self, other):
    from .nn.functional import sub
    return sub(other, self)

  def __rmul__(self, other):
    from .nn.functional import mul
    return mul(other, self)

  def __rtruediv__(self, other):
    from .nn.functional import truediv
    return truediv(other, self)

  def __ge__(self, other):
    from .nn.functional import greater_equal
    return greater_equal(self, other)


class LongTensor(Tensor):
  def __init__(self, *args):
    super(LongTensor, self).__init__(dtype="int64", *args)


class FloatTensor(Tensor):
  def __init__(self, *args):
    super(FloatTensor, self).__init__(dtype="float32", *args)




from ...tensor import Tensor
from .module import Module
from typing import Union, Tuple, Optional
from .utils import _pair, _quadruple, _ntuple
from ..common_types import _size_2_t, _size_4_t, _size_6_t


class GenericPadNd(Module):
  padding: Tuple[int, ...] = None  # set by instance
  mode: str = None  # set by subclass
  value: float = None  # set by subclass instance
  nd: int = None  # set by subclass

  def __init__(self, *, padding: Tuple[int, ...], mode: Optional[str] = None, value: Optional[float] = None):
    super(GenericPadNd, self).__init__()
    self.padding = padding
    if mode is not None:
      self.mode = mode
    if value is not None:
      self.value = value

  def create_returnn_layer_dict(self, input: Tensor):
    assert self.mode
    assert self.mode != "replicate"  # not implemented
    # PyTorch assumes the input to be in batch-feature-major.
    # E.g. for 1D, it assumes input (N, C, W_in),
    # and produces output (N, C, W_out) with W_out = W_in + padding_left + padding_right.
    # For 2D, it assumes input (N, C, H_in, W_in).
    # For 3D, it assumes input (N, C, D_in, H_in, W_in).
    # I.e. does padding in the spatial axes.
    d = {
      "class": "pad", "mode": self.mode, "axes": "spatial", "padding": self.padding,
      "from": self._get_input_layer_name(input)}
    if self.mode == "constant":
      d["value"] = self.value
    return d


class _ConstantPadNd(GenericPadNd):
  mode = "constant"

  def __init__(self, *, padding: Tuple[int, ...], value: float) -> None:
    super(_ConstantPadNd, self).__init__(padding=padding)
    self.value = value


class ConstantPad1d(_ConstantPadNd):
  nd = 1

  def __init__(self, padding: _size_2_t, value: float):
    super(ConstantPad1d, self).__init__(padding=_pair(padding), value=value)


class _ReflectionPadNd(GenericPadNd):
  mode = "reflect"


class ReflectionPad1d(_ReflectionPadNd):
  nd = 1

  def __init__(self, padding: _size_2_t):
    super(ReflectionPad1d, self).__init__(padding=_pair(padding))


class _ReplicationPadNd(GenericPadNd):
  mode = "replicate"


class ReplicationPad1d(_ReplicationPadNd):
  nd = 1

  def __init__(self, padding: _size_2_t) -> None:
    super(ReplicationPad1d, self).__init__(padding=_pair(padding))


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




"""
Dummies...
"""

from __future__ import annotations
from typing import Union, Any, Tuple


def device(name):
  return name


# noinspection PyPep8Naming
class dtype:
  def __init__(self, arg: Union[str, dtype]):
    if isinstance(arg, str):
      self.name = arg
    elif isinstance(arg, dtype):
      self.name = arg.name  # type: str
    else:
      raise TypeError(f"unexpected arg {arg}")

  def __str__(self):
    return f"torch.{self.name}"

  def __eq__(self, other: Union[str, dtype, Any]):
    if isinstance(other, str):
      return self.canonical_name == dtype(other).canonical_name
    if isinstance(other, dtype):
      return self.canonical_name == other.canonical_name
    return False

  def __ne__(self, other: Union[str, dtype, Any]):
    return not (self == other)

  def __hash__(self):
    return hash(self.name)

  @property
  def is_signed(self) -> bool:
    if self.name == "bool":
      return False
    return self.name[:1] != "u"

  @property
  def is_complex(self) -> bool:
    if self.name == "cfloat":
      return True
    if self.name == "cdouble":
      return True
    return self.name.startswith("complex")

  @property
  def is_floating_point(self) -> bool:
    if self.is_complex:
      return True
    if self.name == "double":
      return True
    if self.name == "half":
      return True
    if self.name.startswith("float"):
      return True
    if self.name.startswith("bfloat"):
      return True
    return False

  @property
  def category(self) -> str:
    # https://pytorch.org/docs/stable/tensor_attributes.html
    # complex > floating > integral > boolean
    if self.is_complex:
      return "complex"
    if self.is_floating_point:
      return "float"
    if self.name == "bool":
      return "bool"
    return "int"

  @property
  def category_int(self) -> int:
    return {"bool": 0, "int": 1, "float": 2, "complex": 3}[self.category]

  @property
  def canonical_name(self) -> str:
    """
    Can be used for __eq__.
    """
    d = {
      "half": "float16",
      "float": "float32",
      "double": "float64",
      "cfloat": "complex64",
      "ddouble": "complex128",
      "short": "int16",
      "int": "int32",
      "long": "int64",
    }
    return d.get(self.name, self.name)

  @property
  def bit_size(self) -> int:
    if self.name == "bool":
      return 8
    name = self.canonical_name
    if "128" in name:
      return 128
    if "64" in name:
      return 64
    if "32" in name:
      return 32
    if "16" in name:
      return 16
    if "8" in name:
      return 8
    raise TypeError(f"unexpected dtype {self.name}")


class Size(tuple):  # type: Tuple[SizeValue, ...]
  pass


class SizeValue(int):
  """
  We extend this, to store extra information, e.g. such that this reflects the batch-dim.
  """
  is_batch_dim: bool = False

  def __repr__(self):
    res = super(SizeValue, self).__repr__()
    if self.is_batch_dim:
      res = f"Batch({res})"
    return res


def zeros(*shape):
  from .tensor import Tensor
  return Tensor(*shape)


def from_numpy(arr):
  import numpy
  if isinstance(arr, int):
    arr = numpy.array(arr, dtype='int32')
  if isinstance(arr, float):
    arr = numpy.array(arr, dtype='float32')
  if isinstance(arr, numpy.number):
    arr = numpy.array(arr)
  assert isinstance(arr, numpy.ndarray)
  from .tensor import Tensor
  return Tensor(*arr.shape, dtype=str(arr.dtype), numpy_array=arr)




import torch
from ... import log


# See the discussion here, which is very relevant,
# which covers exactly the topic how one could potentially wrap `torch.Tensor`.
# https://github.com/pytorch/pytorch/blob/master/docs/source/notes/extending.rst#extending-modtorch-with-a-classtensor-like-type

# A soft proxy type unfortunately does not work,
# because `isinstance(obj, torch.Tensor)` checks would not work correctly,
# and they are used in `torch.nn.Module`.


class WrappedTorchTensor(torch.Tensor):
  def __getattribute__(self, item):
    if log.Verbosity >= 10:
      log.unique_print(f"**** torch tensor __getattribute__ {item!r}")
    return super(WrappedTorchTensor, self).__getattribute__(item)

  def new(self, *args, **kwargs):
    # For some reason, new() via __torch_function__ behaves different?
    # This is a workaround.
    if args and isinstance(args[0], torch.Tensor):
      assert len(args) == 1
      assert args[0].type() == self.type()
      return args[0]
    else:
      return self.new_empty(*args, **kwargs)

  @classmethod
  def __torch_function__(cls, func, types, args=(), kwargs=None):
    if log.Verbosity >= 6 and func not in {torch.Tensor.__repr__, torch.Tensor.__str__}:
      log.unique_print(f"**** torch tensor func {func.__name__}")
    if kwargs is None:
      kwargs = {}
    return super().__torch_function__(func, types, args, kwargs)




from ..tensor import Tensor
from ...naming import Naming


class Parameter(Tensor):
  """
  Distinguish from Tensor.
  """
  def __init__(self, *args, **kwargs):
    super(Parameter, self).__init__(*args, **kwargs)
    from returnn.tf.util.data import Data
    naming = Naming.get_instance()
    tensor_entry = naming.register_tensor(self)
    tensor_entry.is_param = True
    tensor_entry.returnn_data = Data(
      name="_unnamed_param", shape=self.shape, dtype=self.dtype.name,
      batch_dim_axis=None, time_dim_axis=None)
    tensor_entry.returnn_axis_from_torch_axis = {i: i for i in range(len(self.shape))}




import types
from .object import WrappedObject
from ..context import WrapCtx


__all__ = ["WrappedModule", "WrappedSourceModule", "WrappedIndirectModule"]


class WrappedModule(types.ModuleType):
  """
  Base class for :class:`WrappedSourceModule` or :class:`WrappedIndirectModule`.
  """
  def __init__(self, *, name: str, orig_mod: types.ModuleType, ctx: WrapCtx):
    super(WrappedModule, self).__init__(name=name)
    self._wrapped__orig_mod = orig_mod
    self._wrapped__ctx = ctx
    assert not isinstance(orig_mod, WrappedModule)

  def __repr__(self):
    return "<%s %s>" % (self.__class__.__name__, self.__name__)


class WrappedSourceModule(WrappedModule):
  """
  Represents a module where the AST was transformed via :class:`AstImportTransformer`.
  """
  def __init__(self, *, source: str, **kwargs):
    super(WrappedSourceModule, self).__init__(**kwargs)
    self._wrapped__source = source


class WrappedIndirectModule(WrappedModule, WrappedObject):
  """
  This can not be transformed on source level,
  so we wrap all attrib access on-the-fly.
  """
  def __init__(self, **kwargs):
    WrappedModule.__init__(self, **kwargs)
    WrappedObject.__init__(self, orig_obj=self._wrapped__orig_mod, name=self.__name__, ctx=self._wrapped__ctx)
    if getattr(self._wrapped__orig_mod, "__all__", None) is not None:
      # noinspection PyUnresolvedReferences
      self.__all__ = self._wrapped__orig_mod.__all__
    else:
      names = sorted(vars(self._wrapped__orig_mod).keys())
      self.__all__ = [name for name in names if name[:1] != "_"]

  def __getattr__(self, item):
    assert item not in {"_wrapped__orig_mod", "__name__"}
    if item == "__path__":  # some special logic, to avoid loading source directly
      if getattr(self._wrapped__orig_mod, "__path__", None) is not None:
        return []
      return None
    return WrappedObject.__getattr__(self, item)

  # setattr not needed
  # but if so, standard keys:
  # {"_wrapped__orig_mod", "__all__", "__loader__", "__package__", "__spec__", "__name__", "__path__"}





from typing import TypeVar, Any
from .. import Module
from ..parameter import Parameter
from ...tensor import Tensor
from ..functional import norm_except_dim


class WeightNorm(object):
  name: str
  dim: int

  def __init__(self, name: str, dim: int) -> None:
    if dim is None:
      dim = -1
    self.name = name
    self.dim = dim

  def compute_weight(self, module: Module) -> Any:
    g = getattr(module, self.name + '_g')
    v = getattr(module, self.name + '_v')
    # or native _weight_norm
    return v * (g / norm_except_dim(v, 2, self.dim))

  @staticmethod
  def apply(module, name: str, dim: int) -> 'WeightNorm':
    if dim is None:
      dim = -1

    fn = WeightNorm(name, dim)

    weight = getattr(module, name)

    # remove w from parameter list
    del module._parameters[name]

    # add g and v as new parameters and express w as g/||v|| * v
    module.register_parameter(name + '_g', Parameter(norm_except_dim(weight, 2, dim).data))
    module.register_parameter(name + '_v', Parameter(weight.data))
    setattr(module, name, fn.compute_weight(module))

    # recompute weight before every forward()
    module.register_forward_pre_hook(fn)

    return fn

  def remove(self, module: Module) -> None:
    weight = self.compute_weight(module)
    delattr(module, self.name)
    del module._parameters[self.name + '_g']
    del module._parameters[self.name + '_v']
    setattr(module, self.name, Parameter(weight.data))

  def __call__(self, module: Module, inputs: Any) -> None:
    setattr(module, self.name, self.compute_weight(module))


T_module = TypeVar('T_module', bound=Module)


def weight_norm(module: T_module, name: str = 'weight', dim: int = 0) -> T_module:
  WeightNorm.apply(module, name, dim)
  return module


def remove_weight_norm(module: T_module, name: str = 'weight') -> T_module:
  for k, hook in module._forward_pre_hooks.items():
    if isinstance(hook, WeightNorm) and hook.name == name:
      hook.remove(module)
      del module._forward_pre_hooks[k]
      return module
  raise ValueError(f"weight_norm of {name!r} not found in {module}")




import collections.abc
from itertools import repeat
from typing import List, Union


def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))
    return parse


_single = _ntuple(1)
_pair = _ntuple(2)
_triple = _ntuple(3)
_quadruple = _ntuple(4)


def _reverse_repeat_tuple(t, n):
    r"""Reverse the order of `t` and repeat each element for `n` times.

    This can be used to translate padding arg used by Conv and Pooling modules
    to the ones used by `F.pad`.
    """
    return tuple(x for x in reversed(t) for _ in range(n))


def _list_with_default(out_size: Union[int, List[int]], defaults: List[int]) -> Union[int, List[int]]:
    if isinstance(out_size, int):
        return out_size
    if len(defaults) <= len(out_size):
        raise ValueError('Input dimension should be at least {}'.format(len(out_size) + 1))
    return [v if v is not None else d for v, d in zip(out_size, defaults[-len(out_size):])]



#!/usr/bin/env python3

import _setup_env  # noqa
import numpy
import better_exchook
from pytorch_to_returnn.converter import verify_torch_and_convert_to_returnn


Batch = 3
Time = 23
Dim = 19


def model_func(wrapped_import, inputs):
  if wrapped_import:
    conformer = wrapped_import("conformer")
  else:
    import conformer

  # https://github.com/lucidrains/conformer
  layer = conformer.ConformerConvModule(
    dim=Dim,
    causal=False,  # auto-regressive or not - 1d conv will be made causal with padding if so
    expansion_factor=2,  # what multiple of the dimension to expand for the depthwise convolution
    kernel_size=31,  # kernel size, 17 - 31 was said to be optimal
    dropout=0.  # dropout at the very end
  )
  layer.eval()  # disable dropout for the test

  x = inputs
  x = layer(x) + x
  return x


def main():
  inputs = numpy.random.randn(Batch, Time, Dim).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=inputs, inputs_data_kwargs={
      "shape": (None, Dim), "time_dim_axis": 1, "feature_dim_axis": 2})


if __name__ == '__main__':
  better_exchook.install()
  main()





from typing import Union, Any
from .base_wrappers.module import WrappedModule
from .context import make_torch_traced_ctx, make_torch_returnn_ctx
from .import_ import import_module


_TorchTracedModPrefix = "%s._torch_traced." % __package__
_wrap_torch_traced_ctx = make_torch_traced_ctx(wrapped_mod_prefix=_TorchTracedModPrefix)

_TorchReturnnModPrefix = "%s._torch_returnn." % __package__
_wrap_torch_returnn_ctx = make_torch_returnn_ctx(wrapped_mod_prefix=_TorchReturnnModPrefix)


def wrapped_import_torch_traced(mod_name: str) -> Union[WrappedModule, Any]:  # rtype Any to make type checkers happy
  """
  :param str mod_name: full mod name, e.g. "custom_torch_code"
  :return: wrapped module
  """
  return import_module(mod_name, ctx=_wrap_torch_traced_ctx)


def wrapped_import_torch_returnn(mod_name: str) -> Union[WrappedModule, Any]:  # rtype Any to make type checkers happy
  """
  :param str mod_name: full mod name, e.g. "custom_torch_code"
  :return: wrapped module
  """
  return import_module(mod_name, ctx=_wrap_torch_returnn_ctx)




import _setup_test_env  # noqa
import sys
import unittest
import typing
import numpy
from pytorch_to_returnn import torch
from pytorch_to_returnn.converter import verify_torch_and_convert_to_returnn


def test_mnist():
  def model_func(wrapped_import, inputs):
    if wrapped_import:
      torch = wrapped_import("torch")
      nn = wrapped_import("torch.nn")
      F = wrapped_import("torch.nn.functional")
    else:
      import torch
      import torch.nn as nn
      import torch.nn.functional as F

    # directly from here: https://github.com/pytorch/examples/blob/master/mnist/main.py
    class Net(nn.Module):
      def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, 3, 1)
        self.conv2 = nn.Conv2d(32, 64, 3, 1)
        self.dropout1 = nn.Dropout(0.25)
        self.dropout2 = nn.Dropout(0.5)
        self.fc1 = nn.Linear(9216, 128)
        self.fc2 = nn.Linear(128, 10)

      def forward(self, x):
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, 2)
        x = self.dropout1(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = F.relu(x)
        x = self.dropout2(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output

    net = Net()
    net = net.eval()  # disable dropout
    return net(inputs)

  rnd = numpy.random.RandomState(42)
  N, C, H, W = 64, 1, 28, 28
  x = rnd.normal(0., 1., (N, C, H, W)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x, inputs_data_kwargs={"shape": (C, H, W)})


def test_custom_layer_norm():
  N, F, T = 64, 11, 28

  def model_func(wrapped_import, inputs):
    if wrapped_import:
      torch = wrapped_import("torch")
    else:
      import torch

    class LayerNorm(torch.nn.LayerNorm):
      def __init__(self, nout, dim=-1):
        """Construct an LayerNorm object."""
        super(LayerNorm, self).__init__(nout, eps=1e-12)
        self.dim = dim

      def forward(self, x):
        if self.dim == -1:
          return super(LayerNorm, self).forward(x)
        return super(LayerNorm, self).forward(x.transpose(1, -1)).transpose(1, -1)

    class Net(torch.nn.Module):
      def __init__(self):
        super(Net, self).__init__()
        self.norm = LayerNorm(nout=F, dim=1)

      def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.norm(x)

    net = Net()
    return net(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (N, F, T)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_naming_problem():
  N, F, T = 3, 5, 11

  def model_func(wrapped_import, inputs):
    if wrapped_import:
      torch = wrapped_import("torch")
    else:
      import torch

    class Activation(torch.nn.Module):
      def forward(self, x):
        return torch.nn.GELU()(x)

    class ConvFeatureExtractionModel(torch.nn.Module):
      def __init__(self):
        super(ConvFeatureExtractionModel, self).__init__()
        self.conv_layers = torch.nn.ModuleList()
        in_d = F
        conv_layers = [(7, 3), (7, 3)]
        for i, (dim, k) in enumerate(conv_layers):
          self.conv_layers.append(
            torch.nn.Sequential(
              torch.nn.Conv1d(in_d, dim, k),
              Activation()))
          in_d = dim

      def forward(self, x):
        for conv in self.conv_layers:
          x = conv(x)
        return x

    class MainModel(torch.nn.Module):
      def __init__(self):
        super(MainModel, self).__init__()
        self.feature_extractor = ConvFeatureExtractionModel()

      def forward(self, x):
        return self.feature_extractor(x)

    model = MainModel()
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (N, F, T)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


if __name__ == "__main__":
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




from typing import Optional, Tuple, Any, List, Dict, Union
from returnn.tf.layers.basic import LayerBase
from .module import Module
from ...tensor import Tensor, dtype as _dtype
from ....naming import Naming, TensorEntry


class Copy(Module):
  is_original_torch_module = False

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "copy", "from": self._get_input_layer_name(input)}

  def make_output_tensor_from_returnn(self, inputs_flat: List[Tensor], layer: LayerBase) -> Tensor:
    assert len(inputs_flat) == 1
    return inputs_flat[0]


class GetSublayer(Module):
  is_original_torch_module = False

  def __init__(self, sub_layer: str):
    super(GetSublayer, self).__init__()
    self.sub_layer = sub_layer

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "copy", "from": f"{self._get_input_layer_name(input)}/{self.sub_layer}"}


class Cast(Module):
  is_original_torch_module = False

  def __init__(self, dtype: Union[str, _dtype]):
    super(Cast, self).__init__()
    self.dtype = _dtype(dtype)

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "cast", "from": self._get_input_layer_name(input), "dtype": self.dtype.name}


class BinaryOperator(Module):
  is_original_torch_module = False

  def __init__(self, kind: str):
    """
    :param kind: "add", "sub", "mul", "truediv"
    """
    super(BinaryOperator, self).__init__()
    self.kind = kind

  def get_returnn_name(self) -> str:
    return self.kind

  def create_returnn_layer_dict(self, *inputs: Tensor):
    inputs = _unify_tensor_dyn_axes(*inputs)
    return {
      "class": "combine", "kind": self.kind,
      "from": [self._get_input_layer_name(input) for input in inputs]}


class ComparisonOperator(BinaryOperator):
  is_original_torch_module = False

  def __init__(self, kind: str):
    """
    :param str kind: e.g. "equal", "greater", "less" or other supported TF comparison ops
    """
    super(ComparisonOperator, self).__init__(kind=kind)

  def create_returnn_layer_dict(self, *inputs: Tensor):
    inputs = _unify_tensor_dyn_axes(*inputs)
    return {
      "class": "compare", "kind": self.kind,
      "from": [self._get_input_layer_name(input) for input in inputs]}


class Reciprocal(Module):
  """
  1/x or 1/max(eps,x)
  """
  is_original_torch_module = False

  def __init__(self, eps: Optional[float] = None):
    super(Reciprocal, self).__init__()
    self.eps = eps

  def create_returnn_layer_dict(self, input: Tensor):
    x = "source(0)"
    if self.eps is not None:
      x = f"maximum_with_identity_grad({x})"
    return {
      "class": "eval", "eval": f"tf_compat.v1.reciprocal({x})",
      "from": self._get_input_layer_name(input)}


class Max(Module):
  is_original_torch_module = False
  pass  # TODO


class ReturnnReinterpretSameSizeAs(Module):
  """
  Used by :func:`_unify_tensor_dyn_axes`.
  """
  is_original_torch_module = False

  def create_returnn_layer_dict(self, input: Tensor, same_as: Tensor) -> Dict[str, Any]:
    return {
      "class": "reinterpret_data",
      "from": self._get_input_layer_name(input),
      "size_base": self._get_input_layer_name(same_as)}


class Transpose(Module):
  """
  Note: The resulting Torch tensor is transposed as expected.
  However, on the RETURNN side, we actually should never need to transpose,
  as we have dimension tags, and all layers should refer to axes by dim tags.
  So on RETURNN side, this is a no-op.
  """
  is_original_torch_module = False

  def __init__(self, perm: Optional[Union[Dict[int, int], Tuple[int, ...], List[int]]]):
    super(Transpose, self).__init__()
    if perm is None:
      self.perm = None
    else:
      if isinstance(perm, (list, tuple)):
        perm = dict(enumerate(perm))
      assert isinstance(perm, dict)
      self.perm = perm  # type: Optional[Dict[int, int]]
      assert len(perm) == len(set(perm.values()))  # should be unique

  def _get_perm(self, input: Tensor) -> Dict[int, int]:
    if self.perm is None:
      dims = range(len(input.shape))
      return {i: j for (i, j) in zip(dims, reversed(dims))}

    def _axis(i):
      assert -len(input.shape) <= i < len(input.shape)
      if i < 0:
        i += len(input.shape)
      assert 0 <= i < len(input.shape)
      return i
    perm = {_axis(i): _axis(j) for (i, j) in self.perm.items()}
    assert len(perm) == len(set(perm.values()))  # should be unique
    return perm

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    # See comment in class docstring. No need to actually transpose.
    return {"class": "copy", "from": self._get_input_layer_name(input)}

  def _get_output_shape_from_returnn(self,
                                     inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    :return: (torch_shape, returnn_axis_from_torch_axis).
      Torch shape how it would have looked when this would be processed within Torch.
      The RETURNN layer.output shape (order of axes) might look different.

    On RETURNN side, this is a no-op.
    But on Torch side, we transpose as expected.
    """
    input, = inputs_flat
    perm = self._get_perm(input)
    target_axes = set(perm)
    source_axes = set(perm.values())
    ndim = len(input.shape)
    rem_target_axes = [i for i in range(ndim) if i not in target_axes]
    rem_source_axes = [i for i in range(ndim) if i not in source_axes]
    assert len(rem_target_axes) == len(rem_source_axes)
    perm.update({i: j for (i, j) in zip(rem_target_axes, rem_source_axes)})
    assert len(perm) == len(set(perm.values())) == ndim

    naming = Naming.get_instance()
    tensor_entry = naming.tensors[input]
    assert isinstance(tensor_entry, TensorEntry)
    assert tensor_entry.returnn_data and tensor_entry.returnn_axis_from_torch_axis
    assert tensor_entry.returnn_data.batch_ndim == ndim

    out_torch_shape = [input.shape[perm[i]] for i in range(ndim)]
    out_returnn_axis_from_torch_axis = {
      perm[i]: j for (i, j) in tensor_entry.returnn_axis_from_torch_axis.items()}
    return tuple(out_torch_shape), out_returnn_axis_from_torch_axis


class Gather(Module):
  """
  Basically x[pos] but in specific dim (axis).
  """
  is_original_torch_module = False

  def __init__(self, dim: int, pos: int):
    super(Gather, self).__init__()
    self.dim = dim
    self.pos = pos

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "gather", "from": self._get_input_layer_name(input),
      "axis": self._get_input_axis_to_returnn(input, axis=self.dim),
      "position": self.pos}


class Slice(Module):
  """
  Slicing of tensors. Wraps RETURNN's SliceLayer.
  """
  is_original_torch_module = False

  def __init__(self, axis: int, start: Optional[int] = None, stop: Optional[int] = None, step: Optional[int] = None):
    super(Slice, self).__init__()
    self.axis = axis
    self.start = start
    self.stop = stop
    self.step = step

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "slice", "axis": self._get_input_axis_to_returnn(input, axis=self.axis),
      "slice_start": self.start, "slice_end": self.stop, "slice_step": self.step,
      "from": self._get_input_layer_name(input)}


class Stack(Module):
  """
  Wraps RETURNN StackLayer.
  """
  is_original_torch_module = False

  def __init__(self, dim: Optional[int] = None):
    super(Stack, self).__init__()
    self.dim = dim

  def create_returnn_layer_dict(self, *inputs: Tensor) -> Dict[str, Any]:
    return {
      "class": "stack", "axis": self.dim,
      "from": [self._get_input_layer_name(x) for x in inputs]}


class Tile(Module):
  """
  Wraps RETURNN TileLayer.
  """
  is_original_torch_module = False

  def __init__(self, multiples: Dict[int, int]):
    super(Tile, self).__init__()
    assert isinstance(multiples, dict)
    self.multiples = multiples

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    multiples = {
      self._get_input_axis_to_returnn(input, axis=axis): multiple
      for axis, multiple in self.multiples.items()}
    return {
      "class": "tile", "multiples": multiples,
      "from": self._get_input_layer_name(input)}


def _unify_tensor_dyn_axes(*inputs: Tensor) -> Tuple[Tensor, ...]:
  """
  You have multiple inputs which can potentially have different dynamic axes (see RETURNN :class:`Data`),
  and this would add ``reinterpret_data`` layers when needed
  to make sure the seq lengths / dim tags are the same.
  """
  if len(inputs) <= 1:
    return inputs
  naming = Naming.get_instance()
  tensors = [naming.tensors[x] for x in inputs]  # type: List[TensorEntry]
  x0 = tensors[0]
  num_dims = max(x.returnn_data.batch_ndim for x in tensors)
  assert all(x.returnn_data.batch_ndim in {0, num_dims} for x in tensors)
  num_spatial_dims = len(x0.returnn_data.get_spatial_batch_axes())
  if num_spatial_dims == 0:
    return inputs
  # Assume same order of spatial axes, but not matter where B/F is.
  spatial_dims = {}  # spatial idx -> (tensor, DimensionTag) (static != 1, or dynamic)
  for i, x in enumerate(list(tensors)):
    if x.returnn_data.batch_ndim == 0:  # scalars are fine
      continue
    x_spatial_axes = x.returnn_data.get_spatial_batch_axes()
    assert len(x_spatial_axes) == num_spatial_dims, f"unexpected Data {x.returnn_data}"
    used_reinterpret = False
    for spatial_idx in range(num_spatial_dims):
      dim_tag = x.returnn_data.get_dim_tag(x_spatial_axes[spatial_idx])
      if dim_tag.dimension == 1:
        continue  # broadcast dim, so always ok. do not add
      if spatial_idx not in spatial_dims:
        spatial_dims[spatial_idx] = (x, dim_tag)
      else:
        prev_x, prev_dim_tag = spatial_dims[spatial_idx]
        if prev_dim_tag is dim_tag:
          pass  # ok
        elif prev_dim_tag.dimension is not None:
          # Static dimension, need to match.
          assert dim_tag.dimension == prev_dim_tag.dimension, f"not matching Data {x.returnn_data}"
        else:
          assert not used_reinterpret  # It should not happen to use this multiple times.
          # This is the case we wanted to catch here.
          # We have two different dynamic axes, which are expected to represent the same seq lengths.
          # So we unify them via reinterpret_data.
          x_ = ReturnnReinterpretSameSizeAs()(x.tensor(), prev_x.tensor())
          x = naming.tensors[x_]
          tensors[i] = x
          used_reinterpret = True
  return tuple([x.tensor() for x in tensors])


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




from __future__ import annotations
import math
import warnings
import numbers
from typing import Dict, Any, Optional, List, Tuple, Union
import tensorflow as tf
from returnn.tf.layers.basic import LayerBase, SubnetworkLayer
from returnn.tf.layers.rec import RecLayer
from .module import Module
from ..parameter import Parameter
from ...tensor import Tensor
from .. import init


class PackedSequence:  # dummy -- not yet implemented...
  pass


def apply_permutation(tensor: Tensor, permutation: Tensor, dim: int = 1) -> Tensor:
  return tensor.index_select(dim, permutation)


class RNNBase(Module):

  def __init__(self, mode: str, input_size: int, hidden_size: int,
               num_layers: int = 1, bias: bool = True, batch_first: bool = False,
               dropout: float = 0., bidirectional: bool = False) -> None:
    super(RNNBase, self).__init__()
    self.mode = mode
    self.input_size = input_size
    self.hidden_size = hidden_size
    self.num_layers = num_layers
    self.bias = bias
    self.batch_first = batch_first
    self.dropout = float(dropout)
    self.bidirectional = bidirectional
    num_directions = 2 if bidirectional else 1

    if not isinstance(dropout, numbers.Number) or not 0 <= dropout <= 1 or isinstance(dropout, bool):
      raise ValueError("dropout should be a number in range [0, 1] "
                       "representing the probability of an element being "
                       "zeroed")
    if dropout > 0 and num_layers == 1:
      warnings.warn("dropout option adds dropout after all but last "
                    "recurrent layer, so non-zero dropout expects "
                    "num_layers greater than 1, but got dropout={} and "
                    "num_layers={}".format(dropout, num_layers))

    if mode == 'LSTM':
      gate_size = 4 * hidden_size
    elif mode == 'GRU':
      gate_size = 3 * hidden_size
    elif mode == 'RNN_TANH':
      gate_size = hidden_size
    elif mode == 'RNN_RELU':
      gate_size = hidden_size
    else:
      raise ValueError("Unrecognized RNN mode: " + mode)

    self._all_weights = []
    for layer in range(num_layers):
      for direction in range(num_directions):
        layer_input_size = input_size if layer == 0 else hidden_size * num_directions

        w_ih = Parameter(Tensor(gate_size, layer_input_size))
        w_hh = Parameter(Tensor(gate_size, hidden_size))
        b_ih = Parameter(Tensor(gate_size))
        # Second bias vector included for CuDNN compatibility. Only one
        # bias vector is needed in standard definition.
        b_hh = Parameter(Tensor(gate_size))
        layer_params = (w_ih, w_hh, b_ih, b_hh)

        suffix = '_reverse' if direction == 1 else ''
        param_names = ['weight_ih_l{}{}', 'weight_hh_l{}{}']
        if bias:
          param_names += ['bias_ih_l{}{}', 'bias_hh_l{}{}']
        param_names = [x.format(layer, suffix) for x in param_names]

        for name, param in zip(param_names, layer_params):
          setattr(self, name, param)
        self._all_weights.append(param_names)

    self.reset_parameters()

  def flatten_parameters(self) -> None:
    pass

  def reset_parameters(self) -> None:
    stdv = 1.0 / math.sqrt(self.hidden_size)
    for weight in self.parameters():
      init.uniform_(weight, -stdv, stdv)

  def check_input(self, input: Tensor, batch_sizes: Optional[Tensor]) -> None:
    expected_input_dim = 2 if batch_sizes is not None else 3
    if input.dim() != expected_input_dim:
      raise RuntimeError(
        'input must have {} dimensions, got {}'.format(
          expected_input_dim, input.dim()))
    if self.input_size != input.size(-1):
      raise RuntimeError(
        'input.size(-1) must be equal to input_size. Expected {}, got {}'.format(
          self.input_size, input.size(-1)))

  def get_expected_hidden_size(self, input: Tensor, batch_sizes: Optional[Tensor]) -> Tuple[int, int, int]:
    if batch_sizes is not None:
      mini_batch = batch_sizes[0]
      mini_batch = int(mini_batch)
    else:
      mini_batch = input.size(0) if self.batch_first else input.size(1)
    num_directions = 2 if self.bidirectional else 1
    expected_hidden_size = (self.num_layers * num_directions,
                            mini_batch, self.hidden_size)
    return expected_hidden_size

  def check_hidden_size(self, hx: Tensor, expected_hidden_size: Tuple[int, int, int],
                        msg: str = 'Expected hidden size {}, got {}') -> None:
    if hx.size() != expected_hidden_size:
      raise RuntimeError(msg.format(expected_hidden_size, list(hx.size())))

  def check_forward_args(self, input: Tensor, hidden: Tensor, batch_sizes: Optional[Tensor]):
    self.check_input(input, batch_sizes)
    expected_hidden_size = self.get_expected_hidden_size(input, batch_sizes)

    self.check_hidden_size(hidden, expected_hidden_size)

  def permute_hidden(self, hx: Tensor, permutation: Optional[Tensor]):
    if permutation is None:
      return hx
    return apply_permutation(hx, permutation)

  def create_returnn_layer_dict(self, input: Tensor, hx: Optional[Tensor] = None) -> Dict[str, Any]:
    assert self.num_layers == 1  # not implemented otherwise
    return {
      "class": "rec", "unit": self.mode, "from": self._get_input_layer_name(input), "n_out": self.hidden_size}

  def import_params_torch_to_returnn(self, *, layer: LayerBase, torch_module):
    pass  # TODO...

  def __setstate__(self, d):
    super(RNNBase, self).__setstate__(d)
    if 'all_weights' in d:
      self._all_weights = d['all_weights']

    if isinstance(self._all_weights[0][0], str):
      return
    num_layers = self.num_layers
    num_directions = 2 if self.bidirectional else 1
    self._all_weights = []
    for layer in range(num_layers):
      for direction in range(num_directions):
        suffix = '_reverse' if direction == 1 else ''
        weights = ['weight_ih_l{}{}', 'weight_hh_l{}{}', 'bias_ih_l{}{}', 'bias_hh_l{}{}']
        weights = [x.format(layer, suffix) for x in weights]
        if self.bias:
          self._all_weights += [weights]
        else:
          self._all_weights += [weights[:2]]

  @property
  def all_weights(self) -> List[Parameter]:
    return [[getattr(self, weight) for weight in weights] for weights in self._all_weights]


class LSTM(RNNBase):
  def __init__(self, *args, **kwargs):
    super(LSTM, self).__init__('LSTM', *args, **kwargs)

  def check_forward_args(self, input: Tensor, hidden: Tuple[Tensor, Tensor], batch_sizes: Optional[Tensor]):
    self.check_input(input, batch_sizes)
    expected_hidden_size = self.get_expected_hidden_size(input, batch_sizes)

    self.check_hidden_size(hidden[0], expected_hidden_size,
                           'Expected hidden[0] size {}, got {}')
    self.check_hidden_size(hidden[1], expected_hidden_size,
                           'Expected hidden[1] size {}, got {}')

  def permute_hidden(self, hx: Tuple[Tensor, Tensor], permutation: Optional[Tensor]) -> Tuple[Tensor, Tensor]:
    if permutation is None:
      return hx
    return apply_permutation(hx[0], permutation), apply_permutation(hx[1], permutation)

  def create_returnn_layer_dict(self, input: Tensor, hx: Optional[Tensor] = None) -> Dict[str, Any]:
    assert not self.bidirectional
    if self.num_layers > 1:
      input_layer_name = self._get_input_layer_name(input)  # call now, to have nicer order of "data"
      subnet_dict = {}
      for i in range(self.num_layers):
        layer_dict = {
          "class": "rec", "unit": "nativelstm2", "from": "data" if i == 0 else f"layer{i - 1}",
          "n_out": self.hidden_size}
        subnet_dict[f"layer{i}"] = layer_dict
        if hx is not None:
          assert isinstance(hx, (list, tuple)) and len(hx) == 2
          h_0, c_0 = hx  # h_0/c_0 of shape (num_layers * num_directions, batch, hidden_size)
          h_, c_ = h_0[i], c_0[i]  # (batch,hidden) each
          layer_dict["initial_state"] = {
            "h": f"base:{self._get_input_layer_name(h_)}",
            "c": f"base:{self._get_input_layer_name(c_)}"}
      subnet_dict["output"] = {"class": "copy", "from": f"layer{self.num_layers - 1}"}
      return {
        "class": "subnetwork", "from": input_layer_name, "subnetwork": subnet_dict}
    d = {
      "class": "rec", "unit": "nativelstm2", "from": self._get_input_layer_name(input),
      "n_out": self.hidden_size}
    if hx is not None:
      assert isinstance(hx, (list, tuple)) and len(hx) == 2
      h_0, c_0 = hx  # h_0/c_0 of shape (num_layers * num_directions, batch, hidden_size)
      h_, c_ = h_0[0], c_0[0]  # (batch,hidden) each
      d["initial_state"] = {
        "h": self._get_input_layer_name(h_),
        "c": self._get_input_layer_name(c_)}
    return d

  def make_structured_returnn_output(self, output: Tensor) -> Union[Tensor, Tuple[Tensor], Any]:
    assert not self.bidirectional
    hs, cs = [], []
    if self.num_layers > 1:
      for i in range(self.num_layers):
        hs.append(GetLastHiddenState(sub_layer=f"layer{i}", key="h")(output))
        cs.append(GetLastHiddenState(sub_layer=f"layer{i}", key="c")(output))
    else:
      hs.append(GetLastHiddenState(key="h")(output))
      cs.append(GetLastHiddenState(key="c")(output))
    from .operator import Stack
    h = Stack(dim=0)(*hs)
    c = Stack(dim=0)(*cs)
    return output, (h, c)

  def check_returnn_layer(self, layer: LayerBase):
    if self.num_layers > 1:
      assert isinstance(layer, SubnetworkLayer)
      assert layer.subnetwork.extern_data.data["data"].dim == self.input_size
    else:
      assert isinstance(layer, RecLayer)
      assert layer.input_data.dim == self.input_size

  def import_params_torch_to_returnn(self, *, layer: LayerBase, torch_module: LSTM):
    import torch
    import numpy
    session = tf.compat.v1.get_default_session()
    for i in range(self.num_layers):
      if self.num_layers > 1:
        assert isinstance(layer, SubnetworkLayer)
        sub_layer = layer.subnetwork.layers[f"layer{i}"]
        assert isinstance(sub_layer, RecLayer)
      else:
        assert isinstance(layer, RecLayer)
        sub_layer = layer
      # i = input_gate, j = new_input, f = forget_gate, o = output_gate
      # RETURNN NativeLstm2 H order: j, i, f, o
      # Torch H order: i, f, j, o  ("W_ii|W_if|W_ig|W_io" in their docs, g==j)
      weight_hh = getattr(torch_module, f"weight_hh_l{i}")  # (H, out)
      weight_ih = getattr(torch_module, f"weight_ih_l{i}")  # (H, in)
      bias_hh = getattr(torch_module, f"bias_hh_l{i}")  # (H,)
      bias_ih = getattr(torch_module, f"bias_ih_l{i}")  # (H,)
      assert isinstance(weight_hh, torch.nn.Parameter)
      assert isinstance(weight_ih, torch.nn.Parameter)
      assert isinstance(bias_hh, torch.nn.Parameter)
      assert isinstance(bias_ih, torch.nn.Parameter)
      bias_np = bias_hh.detach().cpu().numpy() + bias_ih.detach().cpu().numpy()  # RETURNN has single bias
      weight_hh_np = weight_hh.detach().cpu().numpy().transpose()  # (out, H)
      weight_ih_np = weight_ih.detach().cpu().numpy().transpose()  # (in, H)

      def _torch_to_returnn(x: numpy.ndarray, axis: int) -> numpy.ndarray:
        x_ = numpy.split(x, 4, axis=axis)  # (i,f,j,o)
        y = numpy.concatenate([x_[2], x_[0], x_[1], x_[3]], axis=axis)  # -> (j,i,f,o)
        return y

      bias_np = _torch_to_returnn(bias_np, axis=0)
      weight_hh_np = _torch_to_returnn(weight_hh_np, axis=1)
      weight_ih_np = _torch_to_returnn(weight_ih_np, axis=1)

      # RETURNN sub_layer.params: W (in, X), W_re (out, X), b (X,)
      sub_layer.params["W"].load(weight_ih_np, session=session)
      sub_layer.params["W_re"].load(weight_hh_np, session=session)
      sub_layer.params["b"].load(bias_np, session=session)


class GetLastHiddenState(Module):
  is_original_torch_module = False

  def __init__(self, *, sub_layer: Optional[str] = None, key: Optional[str] = None):
    super(GetLastHiddenState, self).__init__()
    self.sub_layer = sub_layer
    self.key = key

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    layer = self._get_input_layer_name(input)
    if self.sub_layer:
      layer = f"{layer}/{self.sub_layer}"
    d = {
      "class": "get_last_hidden_state", "from": layer,
      "n_out": input.shape[-1]}
    if self.key:
      d["key"] = self.key
    return d


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




from __future__ import annotations
from typing import Optional
from returnn.config import Config
from returnn.tf.network import TFNetwork, ExternData
from returnn.tf.layers.basic import LayerBase, SubnetworkLayer
from . import tensor as _tensor


class ReturnnContext:
  def __init__(self, *, parent: Optional[ReturnnContext] = None, name: Optional[str] = None, returnn_train_flag: bool):
    self.parent = parent
    if parent:
      assert name
      self.config = parent.config
      self.tf_name_scope = parent.network.get_absolute_name_scope_prefix() + LayerBase.cls_get_tf_scope_name(name)
      assert parent.network.extern_data.data
      self.sub_net_layer = (
        parent.network.add_layer(
          name=name, layer_class=SubnetworkLayer,
          # This is just a placeholder, will be replaced in define_output.
          sources=[parent.network.get_layer("data")],
          subnetwork={"output": {"class": "copy"}}))  # type: SubnetworkLayer
      self._dummy_sub_output = self.sub_net_layer.subnetwork.layers["output"]
    else:
      self.config = Config({
        # "debug_print_layer_output_template": True,
      })
      self.tf_name_scope = ""
      self.sub_net_layer = None
      self._dummy_sub_output = None
    if self.sub_net_layer:
      self.network = self.sub_net_layer.subnetwork
    else:
      assert not parent
      self.network = TFNetwork(
        extern_data=ExternData(), config=self.config, name="root",
        train_flag=returnn_train_flag,
        absolute_name_prefix=(self.tf_name_scope + "/") if self.tf_name_scope else "")

  def __repr__(self):
    return f"<{self.__class__.__name__} {self.network.get_absolute_name_prefix()!r}>"

  def define_input(self, input: _tensor.TensorEntry, *, data_key: Optional[str] = None):
    if self._dummy_sub_output:
      assert self.network.layers["output"] is self._dummy_sub_output
      self._dummy_sub_output = None
      # Reset both, as we refill them. They contain dummy data.
      self.network.layers.clear()
      self.network.extern_data.data.clear()
    if data_key is None:
      data_key = self.network.extern_data.default_input
    assert data_key not in self.network.extern_data.data
    assert input.returnn_data
    self.network.extern_data.data[data_key] = input.returnn_data

  def define_output(self, output_layer: LayerBase):
    assert self.network.layers["output"] is output_layer
    if self.sub_net_layer:
      self.sub_net_layer.output = output_layer.output



#!/usr/bin/env python3

"""
Usage example::

  python3 decoder.py \
  --returnn_config tacotron2_ljspeech.config.py \
  --vocab_file data/cmu_vocab.pkl \
  --pronunciation_lexicon data/cmudict.dict \
  --pwg_config mb_melgan.v2.yaml \
  --pwg_checkpoint mb_melgan_models/checkpoint-1000000steps.pkl \
  --input "hello world"

"""


import argparse
from num2words import num2words
import numpy
import os
import subprocess
import sys
import torch
import yaml
import wave
import better_exchook

from parallel_wavegan import models as pwg_models
from parallel_wavegan.layers import PQMF
from returnn import __main__ as rnn
from returnn.datasets.generating import StaticDataset, Vocabulary


def number_convert(word):
    try:
        f = float(word)
        return num2words(f)
    except Exception:
        return word


def main():
    parser = argparse.ArgumentParser(description="TTS decoder running RETURNN TTS and an MB-MelGAN vocoder")
    parser.add_argument("--input", help="use this input (otherwise stdin as default)")
    parser.add_argument("--returnn_config", type=str, help="RETURNN config file (.config)")
    parser.add_argument("--vocab_file", type=str, help="RETURNN vocab file (.pkl)")
    parser.add_argument("--pronunciation_lexicon", type=str, help="CMU style pronuncation lexicon")
    parser.add_argument("--pwg_config", type=str, help="ParallelWaveGAN config (.yaml)")
    parser.add_argument("--pwg_checkpoint", type=str, help="ParallelWaveGAN checkpoint (.pkl)")
    parser.add_argument("--dump_features", help="npy file")

    args = parser.parse_args()

    # Initialize RETURNN
    rnn.init(args.returnn_config, config_updates={"train": None, "dev": None, "device": "cpu"})

    rnn.engine.use_search_flag = True # enable search mode
    rnn.engine.init_network_from_config(rnn.config)

    returnn_vocab = Vocabulary(vocab_file=args.vocab_file, unknown_label=None)
    returnn_output_dict = {'output': rnn.engine.network.get_default_output_layer().output.placeholder}

    # Initialize PWG
    pwg_config = yaml.load(open(args.pwg_config), Loader=yaml.Loader)
    pyt_device = torch.device("cpu")
    generator = pwg_models.MelGANGenerator(**pwg_config['generator_params'])
    generator.load_state_dict(
        torch.load(args.pwg_checkpoint, map_location="cpu")["model"]["generator"])
    generator.remove_weight_norm()
    pwg_model = generator.eval().to(pyt_device)
    pwg_pad_fn = torch.nn.ReplicationPad1d(
        pwg_config["generator_params"].get("aux_context_window", 0))
    pwg_pqmf = PQMF(pwg_config["generator_params"]["out_channels"]).to(pyt_device)

    # load a CMU dict style pronunciation table
    pronunciation_dictionary = {}
    with open(args.pronunciation_lexicon, "rt") as lexicon:
        for lexicon_entry in lexicon.readlines():
            word, phonemes = lexicon_entry.strip().split(" ", maxsplit=1)
            pronunciation_dictionary[word] = phonemes.split(" ")

    def gen_audio(line: str) -> bytes:
        line = line.strip().lower()

        # Tokenizer perl command
        tokenizer = ["perl", "./scripts/tokenizer/tokenizer.perl", "-l", "en", "-no-escape"]
        # run perl tokenizer as external script
        p = subprocess.Popen(tokenizer, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        line = p.communicate(input=line.encode("UTF-8"))[0].decode("UTF-8").strip()
        p.terminate()
        print("Tokenized: %r" % line)
        assert line

        # apply num2wordsn and pronunciation dict
        words = list(map(number_convert, line.split()))
        print("Words:", words)
        phoneme_sequence = " _ ".join(
            [" ".join(pronunciation_dictionary[w]) for w in words if w in pronunciation_dictionary.keys()])
        phoneme_sequence += " _ ~"
        print("Phonemes:", phoneme_sequence)

        try:
            classes = numpy.asarray(returnn_vocab.get_seq(phoneme_sequence), dtype="int32")
            feed_dict = {'classes': classes}
            dataset = StaticDataset([feed_dict], output_dim={'classes': (77, 1)})
            result = rnn.engine.run_single(dataset, 0, returnn_output_dict)
        except Exception as e:
            print(e)
            raise e

        feature_data = numpy.squeeze(result['output']).T
        print("Feature shape:", feature_data.shape)
        if args.dump_features:
            print("Dump features to file:", args.dump_features)
            numpy.save(args.dump_features, feature_data)

        with torch.no_grad():
            input_features = pwg_pad_fn(torch.from_numpy(feature_data).unsqueeze(0)).to(pyt_device)
            audio_waveform = pwg_pqmf.synthesis(pwg_model(input_features)).view(-1).cpu().numpy()

        return numpy.asarray(audio_waveform*(2**15-1), dtype="int16").tobytes()

    audios = []
    if args.input is not None:
        audios.append(gen_audio(args.input))
    else:
        print("Reading from stdin.")
        for line in sys.stdin.readlines():
            if not line.strip():
                continue
            audios.append(gen_audio(line))

    for i, audio in enumerate(audios):
        out_fn = "out_%i.wav" % i
        wave_writer = wave.open(out_fn, "wb")
        wave_writer.setnchannels(1)  
        wave_writer.setframerate(16000)
        wave_writer.setsampwidth(2)
        wave_writer.writeframes(audio)
        wave_writer.close()
        print("Wrote %s." % out_fn)


if __name__ == "__main__":
    better_exchook.install()
    main()




from .module import Module
from .container import *
from .conv import *
from .pooling import *
from .padding import *
from .activation import *
from .linear import *
from .sparse import *
from .dropout import *
from .batchnorm import *
from .normalization import *
from .rnn import *
from .shape import *
from .norm import *
from .loss import *
from .operator import *
from .variable import *
from .container import __all__ as _container_all
from .conv import __all__ as _conv_all
from .pooling import __all__ as _pooling_all
from .padding import __all__ as _padding_all
from .activation import __all__ as _activation_all
from .linear import __all__ as _linear_all
from .sparse import __all__ as _sparse_all
from .dropout import __all__ as _dropout_all
from .batchnorm import __all__ as _batchnorm_all
from .normalization import __all__ as _normalization_all
from .rnn import __all__ as _rnn_all
from .shape import __all__ as _shape_all
from .norm import __all__ as _norm_all
from .loss import __all__ as _loss_all
from .operator import __all__ as _operator_all
from .variable import __all__ as _variable_all


__all__ = (
    ["Module"] +
    _container_all + _conv_all + _pooling_all + _padding_all +
    _activation_all + _linear_all + _sparse_all + _dropout_all +
    _batchnorm_all + _normalization_all + _rnn_all +
    _shape_all + _norm_all + _loss_all +
    _operator_all + _variable_all)




from ..context import WrapCtx


def copy_attribs_qualname_and_co(target, source, *, ctx: WrapCtx):
  target.__name__ = source.__name__
  target.__qualname__ = source.__qualname__
  if source.__module__:
    if ctx.should_wrap_mod(source.__module__):
      target.__module__ = ctx.mod_map.map_mod_name(source.__module__)
    else:
      target.__module__ = source.__module__




import _setup_test_env  # noqa
import sys
import unittest
import typing
import numpy
from pytorch_to_returnn.pprint import pprint, pformat, _type_simplicity_limit, _type_simplicity_score
from pprint import pprint as orig_pprint


def assert_equal(a_obj, b):
  a = pformat(a_obj)
  a = a.strip()
  b = b.strip()
  print("Repr:", repr(a_obj))
  print("Orig pprint:")
  orig_pprint(a_obj)
  print("New pprint:")
  print(a)
  if a != b:
    print("!=")
    print(b)
  assert a == b


class Obj:
  """
  Custom object to enforce to fallback to generic repr.
  """
  def __init__(self, value):
    self.value = value

  def __repr__(self):
    return f"{self.__class__.__name__}({self.value!r})"


def test_pprint_simple():
  assert_equal({}, "{}")
  assert_equal((), "()")
  assert_equal([], "[]")
  assert_equal(set(), "set()")


def test_pprint_simple_in_single_line():
  assert_equal({1}, "{1}")
  assert_equal({1, 2, 3}, "{1, 2, 3}")


def test_pprint_multi_line():
  assert_equal([Obj(1), 2, 3], """
[
  Obj(1),
  2,
  3
]""")
  assert_equal({"a": Obj(3)}, """
{
  'a': Obj(3)
}""")
  assert_equal({"a": Obj(3), "b": 4}, """
{
  'a': Obj(3),
  'b': 4
}""")
  net_dict_example = {
    'melgan': {
      'class': 'subnetwork',
      'from': 'data',
      'subnetwork': {
        'l0': {'class': 'pad', 'mode': 'reflect', 'axes': 'spatial', 'padding': (3, 3), 'from': 'data'},
        'la1': {
          'class': 'conv',
          'from': 'l0',
          'activation': None,
          'with_bias': True,
          'n_out': 384,
          'filter_size': (7,),
          'padding': 'valid',
          'strides': (1,),
          'dilation_rate': (1,)
        },
        'lay2': {'class': 'eval', 'eval': 'tf.nn.leaky_relu(source(0), alpha=0.2)', 'from': 'la1'},
        'layer3_xxx': {
          'class': 'transposed_conv',
          'from': 'lay2',
          'activation': None,
          'with_bias': True,
          'n_out': 192,
          'filter_size': (10,),
          'strides': (5,),
          'padding': 'valid',
          'output_padding': (1,),
          'remove_padding': (3,)
        },
        'output': {'class': 'copy', 'from': 'layer3_xxx'}
      }
    },
    'output': {'class': 'copy', 'from': 'melgan'}
  }
  assert_equal(net_dict_example, """
{
  'melgan': {
    'class': 'subnetwork',
    'from': 'data',
    'subnetwork': {
      'l0': {'class': 'pad', 'mode': 'reflect', 'axes': 'spatial', 'padding': (3, 3), 'from': 'data'},
      'la1': {
        'class': 'conv',
        'from': 'l0',
        'activation': None,
        'with_bias': True,
        'n_out': 384,
        'filter_size': (7,),
        'padding': 'valid',
        'strides': (1,),
        'dilation_rate': (1,)
      },
      'lay2': {'class': 'eval', 'eval': 'tf.nn.leaky_relu(source(0), alpha=0.2)', 'from': 'la1'},
      'layer3_xxx': {
        'class': 'transposed_conv',
        'from': 'lay2',
        'activation': None,
        'with_bias': True,
        'n_out': 192,
        'filter_size': (10,),
        'strides': (5,),
        'padding': 'valid',
        'output_padding': (1,),
        'remove_padding': (3,)
      },
      'output': {'class': 'copy', 'from': 'layer3_xxx'}
    }
  },
  'output': {'class': 'copy', 'from': 'melgan'}
}""")


def test_pprint_numpy_ndarray():
  assert_equal(numpy.array([[1, 2], [3, 4]]), "numpy.array([[1, 2], [3, 4]], dtype=numpy.int64)")


def test_pprint_dict_numpy_ndarray():
  assert_equal(
    {"a": 42, "b": numpy.array([[1, 2, 3, 4]] * 10)}, """
{
  'a': 42,
  'b': numpy.array([
""" +
"""    [1, 2, 3, 4],
""" * 9 +
"""    [1, 2, 3, 4]
  ], dtype=numpy.int64)
}
""")


if __name__ == "__main__":
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




from typing import Dict


class ModMap:
  def __init__(self, base_mod_map: Dict[str, str]):
    """
    :param base_mod_map: e.g.
      {"parallel_wavegan": "pytorch_to_returnn._wrapped_mods.parallel_wavegan",
       "torch": "pytorch_to_returnn._wrapped_mods.torch"}
      or
      {"parallel_wavegan": "pytorch_to_returnn._demo_mods.parallel_wavegan",
       "torch": "pytorch_to_returnn.torch"}
    """
    self.base_mod_map = base_mod_map.copy()

  def should_wrap_mod_name(self, mod_name: str) -> bool:
    """
    :param mod_name: e.g. "torch.nn.init"
    :return: whether it should be wrapped / can be mapped. see :func:`map_mod_name`
    """
    if mod_name in self.base_mod_map:
      return True
    for base_mod_name in self.base_mod_map.keys():
      if mod_name.startswith(base_mod_name + "."):
        return True
    return False

  def find_base_mod_prefix(self, mod_name: str) -> str:  # e.g. "torch.nn.init" -> "torch"
    """
    :param mod_name: e.g. "torch.nn.init". :func:`should_wrap_mod_name` must return True
    :return: e.g. "torch"
    """
    if mod_name in self.base_mod_map:
      return mod_name
    for base_mod_name in sorted(self.base_mod_map.keys(), reverse=True):
      if mod_name.startswith(base_mod_name + "."):
        return base_mod_name
    raise Exception(f"cannot map mod name {mod_name!r}. map {self.base_mod_map!r}")

  def map_mod_name(self, mod_name: str) -> str:  # e.g. "torch.nn" -> "pytorch_to_returnn._wrapped_mods.torch.nn"
    """
    :param mod_name: e.g. "torch.nn". :func:`should_wrap_mod_name` must return True
    :return: e.g. "pytorch_to_returnn._wrapped_mods.torch.nn"
    """
    base_mod_name = self.find_base_mod_prefix(mod_name)
    postfix = mod_name[len(base_mod_name):]
    assert postfix.startswith(".") or postfix == ""
    return self.base_mod_map[base_mod_name] + postfix

  def simplify_(self):
    """
    Simplify the internal dict without changing semantics.
    This is done inplace.
    """
    for base_mod_name in sorted(self.base_mod_map.keys(), reverse=True):
      mapped_mod_name = self.base_mod_map.pop(base_mod_name)  # just temporarily
      if not self.should_wrap_mod_name(base_mod_name):
        # Need it, put back in.
        self.base_mod_map[base_mod_name] = mapped_mod_name
      else:
        # Some other entry maps.
        mapped_mod_name_ = self.map_mod_name(base_mod_name)
        if mapped_mod_name != mapped_mod_name_:
          # Need it, put back in.
          self.base_mod_map[base_mod_name] = mapped_mod_name
        else:
          pass  # Can keep it that way.




from ... import log
from .object import WrappedObject
from ..context import WrapCtx
from .utils import copy_attribs_qualname_and_co


class WrappedClassBase(WrappedObject):
  pass


def make_wrapped_class(cls: type, name: str, ctx: WrapCtx):
  from ..wrap import wrap, unwrap

  class WrappedClass(WrappedClassBase):
    def __init__(self, *args, **kwargs):
      if log.Verbosity >= 4:
        log.unique_print("*** WrappedClass %s(...)" % (name,))
      if isinstance(kwargs.get("orig_obj"), cls):
        assert not args
        WrappedObject.__init__(self, **kwargs)
      else:
        obj = cls(*unwrap(args), **unwrap(kwargs))
        WrappedObject.__init__(self, orig_obj=obj, name="%s(...)" % name, ctx=ctx)

    # Use __getattribute__ because we might have added `cls` as a base,
    # and want to catch all getattr, also to existing methods.
    def __getattribute__(self, item):
      # if item == "__dict__":
        # Special case. Directly get it from wrapped object.  # TODO use this?
        #  return self._wrapped__orig_obj.__dict__
      if item in {"_wrapped__orig_obj", "_wrapped__name", "_wrapped__ctx", "__class__"}:  # extra checks, and fast path
        return object.__getattribute__(self, item)
      if self._wrapped__orig_obj is self:  # special case
        res = object.__getattribute__(self, item)
        res = wrap(res, name="%s.%s" % (self._wrapped__name, item), ctx=ctx)
        return res
      return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
      if key in {"_wrapped__orig_obj", "_wrapped__name", "_wrapped__ctx"}:
        return object.__setattr__(self, key, value)
      if self is self._wrapped__orig_obj:  # special case
        return object.__setattr__(self, key, value)
      return setattr(self._wrapped__orig_obj, key, value)

  copy_attribs_qualname_and_co(WrappedClass, cls, ctx=ctx)
  try:
    # We add this such that `isinstance(WrappedClass(), cls)` returns True.
    # Note that this could cause some problems (that's why we use __getattribute__ above),
    # and also, it does not always work (e.g. for native types like torch.Tensor).
    WrappedClass.__bases__ += (cls,)
  except TypeError:  # e.g. object layout differs or so
    pass  # just ignore
  return WrappedClass




from typing import Any, Iterable, Iterator, Mapping, Optional, overload, Tuple, TypeVar, Union, Dict
from collections import OrderedDict
from collections import abc as container_abcs
import operator
from itertools import islice
from .module import Module
from ...tensor import Tensor


T = TypeVar('T')


class Sequential(Module):
  @overload
  def __init__(self, *args: Module) -> None:
    ...

  @overload
  def __init__(self, arg: 'OrderedDict[str, Module]') -> None:
    ...

  def __init__(self, *args: Any):
    super(Sequential, self).__init__()
    if len(args) == 1 and isinstance(args[0], OrderedDict):
      for key, module in args[0].items():
        self.add_module(key, module)
    else:
      for idx, module in enumerate(args):
        self.add_module(str(idx), module)

  def _get_item_by_idx(self, iterator, idx):
    """Get the idx-th item of the iterator"""
    size = len(self)
    idx = operator.index(idx)
    if not -size <= idx < size:
      raise IndexError('index {} is out of range'.format(idx))
    idx %= size
    return next(islice(iterator, idx, None))

  def __getitem__(self, idx: Union[slice, int]):
    if isinstance(idx, slice):
      return self.__class__(OrderedDict(list(self._modules.items())[idx]))
    else:
      return self._get_item_by_idx(self._modules.values(), idx)

  def __setitem__(self, idx: int, module: Module) -> None:
    key = self._get_item_by_idx(self._modules.keys(), idx)
    return setattr(self, key, module)

  def __delitem__(self, idx: Union[slice, int]) -> None:
    if isinstance(idx, slice):
      for key in list(self._modules.keys())[idx]:
        delattr(self, key)
    else:
      key = self._get_item_by_idx(self._modules.keys(), idx)
      delattr(self, key)

  def __len__(self) -> int:
    return len(self._modules)

  def __iter__(self) -> Iterator[Module]:
    return iter(self._modules.values())

  def forward(self, input):
    for module in self:
      res = module(input)
      assert isinstance(res, Tensor)
      input = res
    return input


class ModuleList(Module):
  def __init__(self, modules: Optional[Iterable[Module]] = None) -> None:
    super(ModuleList, self).__init__()
    if modules is not None:
      self += modules

  def _get_abs_string_index(self, idx):
    """Get the absolute index for the list of modules"""
    idx = operator.index(idx)
    if not (-len(self) <= idx < len(self)):
      raise IndexError('index {} is out of range'.format(idx))
    if idx < 0:
      idx += len(self)
    return str(idx)

  def __getitem__(self, idx: int) -> Module:
    if isinstance(idx, slice):
      return self.__class__(list(self._modules.values())[idx])
    else:
      return self._modules[self._get_abs_string_index(idx)]

  def __setitem__(self, idx: int, module: Module) -> None:
    idx = self._get_abs_string_index(idx)
    return setattr(self, str(idx), module)

  def __delitem__(self, idx: Union[int, slice]) -> None:
    if isinstance(idx, slice):
      for k in range(len(self._modules))[idx]:
        delattr(self, str(k))
    else:
      delattr(self, self._get_abs_string_index(idx))
    # To preserve numbering, self._modules is being reconstructed with modules after deletion
    str_indices = [str(i) for i in range(len(self._modules))]
    self._modules = OrderedDict(list(zip(str_indices, self._modules.values())))

  def __len__(self) -> int:
    return len(self._modules)

  def __iter__(self) -> Iterator[Module]:
    return iter(self._modules.values())

  def __iadd__(self: T, modules: Iterable[Module]) -> T:
    return self.extend(modules)

  def __dir__(self):
    keys = super(ModuleList, self).__dir__()
    keys = [key for key in keys if not key.isdigit()]
    return keys

  def insert(self, index: int, module: Module) -> None:
    """Insert a given module before a given index in the list."""
    for i in range(len(self._modules), index, -1):
      self._modules[str(i)] = self._modules[str(i - 1)]
    self._modules[str(index)] = module

  def append(self: T, module: Module) -> T:
    """Appends a given module to the end of the list."""
    self.add_module(str(len(self)), module)
    return self

  def extend(self: T, modules: Iterable[Module]) -> T:
    """Appends modules from a Python iterable to the end of the list."""
    if not isinstance(modules, container_abcs.Iterable):
      raise TypeError("ModuleList.extend should be called with an "
                      "iterable, but got " + type(modules).__name__)
    offset = len(self)
    for i, module in enumerate(modules):
      self.add_module(str(offset + i), module)
    return self

  def forward(self):
    raise NotImplementedError()


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




"""
This wraps both torch.nn.functional and torch.functional, and other __torch_function__ on tensors.

Note that this all maps to modules, which will be temporarily created.
In RETURNN, every operation is a layer.
"""

import numpy
from typing import Optional, Union, List, Tuple, Dict, TypeVar, Sequence
from . import modules
from ..tensor import Tensor
from .._C import Size, dtype as _dtype


_number = Union[int, float, numpy.ndarray, numpy.number]
_size = Union[Size, List[int], Tuple[int, ...]]
_T = TypeVar("_T")
_default_float_type = "float32"


def zeros(*size, out=None, dtype=None, layout=None, device=None, requires_grad=False):
  return Tensor(*size, dtype=dtype)


def ones(*size, out=None, dtype=None, layout=None, device=None, requires_grad=False):
  if size and isinstance(size[0], (tuple, list)):
    assert len(size) == 1
    size = tuple(size[0])
  if not dtype:
    dtype = _default_float_type
  return tensor(numpy.ones(size, dtype=dtype))


def tensor(data, *, dtype=None, device=None, requires_grad=False, pin_memory=False):
  from .._C import from_numpy
  x = from_numpy(data)
  if dtype:
    x = x.type(dtype)
  return x


def cast(input: Union[_T, Tensor, _number], dtype: Union[str, _dtype]) -> Union[_T, Tensor]:
  dtype = _dtype(dtype)
  if dtype == get_dtype(input):
    return input
  return modules.Cast(dtype=dtype)(input)


def get_dtype(tensor: Union[Tensor, _number]) -> _dtype:
  if isinstance(tensor, Tensor):
    return tensor.dtype
  if isinstance(tensor, int):
    return _dtype("int32")
  if isinstance(tensor, float):
    return _dtype(_default_float_type)
  if isinstance(tensor, (numpy.number, numpy.ndarray)):
    return _dtype(str(tensor.dtype))
  raise TypeError(f"unexpected type {type(tensor)}")


def result_type(tensor1: Union[Tensor, _number], tensor2: Union[Tensor, _number]) -> _dtype:
  # https://pytorch.org/docs/stable/generated/torch.result_type.html
  return promote_types(get_dtype(tensor1), get_dtype(tensor2))


def promote_types(type1: Union[str, _dtype], type2: Union[str, _dtype]) -> _dtype:
  # https://pytorch.org/docs/stable/generated/torch.promote_types.html
  type1 = _dtype(type1)
  type2 = _dtype(type2)
  if type1.category_int != type2.category_int:
    if type1.category_int < type2.category_int:
      type1, type2 = type2, type1
    assert type1.category_int > type2.category_int
    return type1
  assert type1.category_int == type2.category_int
  if type1.bit_size == type2.bit_size:
    assert type1 == type2
    return type1
  if type1.bit_size < type2.bit_size:
    type1, type2 = type2, type1
  assert type1.bit_size > type2.bit_size
  return type1


def as_tensor(data: Union[Tensor, _number],
              dtype: Optional[Union[str, _dtype]] = None,
              device=None) -> Tensor:
  if not isinstance(data, Tensor):
    from .._C import from_numpy
    data = from_numpy(data)
  assert isinstance(data, Tensor)
  if dtype:
    data = cast(data, dtype)
  return data


def add(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.BinaryOperator(kind="add")(cast(x, dtype), cast(y, dtype))


def sub(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.BinaryOperator(kind="sub")(cast(x, dtype), cast(y, dtype))


def mul(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.BinaryOperator(kind="mul")(cast(x, dtype), cast(y, dtype))


def truediv(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.BinaryOperator(kind="truediv")(cast(x, dtype), cast(y, dtype))


def greater_equal(x: Tensor, y: Tensor) -> Tensor:
  dtype = result_type(x, y)
  return modules.ComparisonOperator(kind="greater_equal")(cast(x, dtype), cast(y, dtype))


def flatten(input: Tensor, start_dim=0, end_dim=-1) -> Tensor:
  return modules.Flatten(start_dim=start_dim, end_dim=end_dim).as_returnn_torch_functional()(input)


def reshape(input: Tensor, shape: Tuple[int, ...]) -> Tensor:
  if any(dim == -1 for dim in shape):
    num = input.numel()
    for dim in shape:
      if dim == -1:
        continue
      assert dim > 0 and num % dim == 0
      num //= dim
    shape = [dim if dim >= 0 else num for dim in shape]

  # Use Flatten, Unflatten, Squeeze.
  # (Other reshapes are disallowed.)
  axis1, axis2 = 0, 0
  while axis1 < len(input.shape) and axis2 < len(shape):
    if input.shape[axis1] == shape[axis2]:
      axis1 += 1
      axis2 += 1
      continue
    elif input.shape[axis1] < shape[axis2]:
      if input.shape[axis1] == 1:
        input = modules.Squeeze(dim=axis1).as_returnn_torch_functional()(input)
        continue
      n = 1
      a = axis1
      while a < len(input.shape) and n < shape[axis2]:
        assert shape[axis2] % n == 0 and n < shape[axis2]
        n *= input.shape[a]
        a += 1
      assert n == shape[axis2]
      input = modules.Flatten(start_dim=axis1, end_dim=a - 1).as_returnn_torch_functional()(input)
      assert input.shape[axis1] == shape[axis2]
      continue
    elif input.shape[axis1] > shape[axis2]:
      n = 1
      a = axis2
      while a < len(shape) and n < input.shape[axis1]:
        assert input.shape[axis1] % n == 0 and n < input.shape[axis1]
        n *= shape[a]
        a += 1
      assert n == input.shape[axis1]
      input = modules.Unflatten(dim=axis1, unflattened_size=tuple(shape[axis2:a])).as_returnn_torch_functional()(input)
      assert input.shape[axis1] == shape[axis2]
      continue
    else:
      assert False  # cannot happen
  assert axis1 == axis2
  if len(input.shape) < len(shape):
    assert all(shape[i] == 1 for i in range(len(input.shape), len(shape)))
    input = modules.Unflatten(
      dim=-1, unflattened_size=shape[len(input.shape) - 1:]).as_returnn_torch_functional()(input)
  elif len(input.shape) > len(shape):
    while len(input.shape) > len(shape):
      input = modules.Squeeze(dim=len(shape)).as_returnn_torch_functional()(input)
  assert len(input.shape) == len(shape) and input.shape == tuple(shape)
  return input


def movedim(input: Tensor, source: Union[int, Tuple[int, ...]], destination: Union[int, Tuple[int, ...]]):
  if isinstance(source, int):
    source = (source,)
  if isinstance(destination, int):
    destination = (destination,)
  assert isinstance(source, (tuple, list)) and isinstance(destination, (tuple, list))
  assert len(source) == len(destination)
  perm = {i: j for i, j in zip(destination, source)}
  # All remaining axes stay in order.
  return tensorflow_transpose(input, perm=perm)


def transpose(input: Tensor, dim0: int, dim1: int):
  return tensorflow_transpose(input, perm={dim0: dim1, dim1: dim0})


def tensorflow_transpose(input: Tensor, perm: Optional[Union[Dict[int, int], Tuple[int, ...], List[int]]]):
  """
  Note: This function is added by us, not available in original PyTorch.

  Note: The resulting Torch tensor is transposed as expected.
  However, on the RETURNN side, we actually should never need to transpose,
  as we have dimension tags, and all layers should refer to axes by dim tags.
  So on RETURNN side, this is a no-op.
  """
  return modules.Transpose(perm=perm)(input)


def expand(input: Tensor, *sizes: _size):
  if sizes and isinstance(sizes[0], (tuple, list)):
    assert len(sizes) == 1
    shape = tuple(sizes[0])
  else:
    shape = sizes
  assert isinstance(shape, tuple)

  # reshape to add new dims
  if len(shape) != len(input.shape):
    assert len(shape) >= len(input.shape), "Cannot reduce dimensions"
    assert -1 not in [d for d in shape[:-len(input.shape)]], "Size -1 is not allowed for expanded dims"
    exp_shape = (1,) * (len(shape) - len(input.shape)) + input.shape[-len(input.shape):]
    input = reshape(input, shape=exp_shape)

  # tile to expand to larger size
  multiples = {}
  for axis, multiple in enumerate(shape):
    if input.shape[axis] > 1:
      assert multiple == -1 or multiple == input.shape[axis], \
        f"Expanded size ({multiple}) must match existing size ({input.shape[axis]}) at non-singleton dimension ({axis})"
      multiple = 1
    if multiple == -1:
      multiple = 1  # -1 as the size for a dim means not changing the size of that dim
    multiples[axis] = multiple
  return modules.Tile(multiples=multiples).as_returnn_torch_functional()(input)


def chunk(input: Tensor, chunks: int, dim: int = 0) -> List[Tensor]:
  chunk_size = input.shape[dim] // chunks
  if input.shape[dim] % chunks != 0:
    chunk_size += 1
    size_splits = [chunk_size] * chunks
    size_splits[-1] -= chunk_size * chunks - input.shape[dim]
  else:
    size_splits = [chunk_size] * chunks
  assert sum(size_splits) == input.shape[dim] and len(size_splits) == chunks
  return modules.Split(dim=dim, size_splits=size_splits).as_returnn_torch_functional()(input)


def pad(input: Tensor, pad, mode='constant', value=0) -> Tensor:
  return modules.GenericPadNd(padding=pad, mode=mode, value=value).as_returnn_torch_functional()(input)


def max(*inputs: Tensor) -> Tensor:
  return modules.Max()(*inputs)


def conv1d(
    input: Tensor, weight: Tensor, bias: Optional[Tensor] = None,
    stride: Union[int, _size] = 1, padding: Union[int, _size] = 0,
    dilation: Union[int, _size] = 1, groups: int = 1) -> Tensor:
  mod = modules.FunctionalConv1d(stride=stride, padding=padding, dilation=dilation, groups=groups)
  return mod(input, weight, bias)


def conv2d(
    input: Tensor, weight: Tensor, bias: Optional[Tensor] = None,
    stride: Union[int, _size] = 1, padding: Union[int, _size] = 0,
    dilation: Union[int, _size] = 1, groups: int = 1) -> Tensor:
  mod = modules.FunctionalConv2d(stride=stride, padding=padding, dilation=dilation, groups=groups)
  return mod(input, weight, bias)


def conv_transpose1d(
    input: Tensor, weight: Tensor, bias: Optional[Tensor] = None,
    stride: Union[int, _size] = 1, padding: Union[int, _size] = 0,
    output_padding: Union[int, _size] = 0,
    groups: int = 1,
    dilation: Union[int, _size] = 1) -> Tensor:
  mod = modules.FunctionalConvTransposed1d(
    stride=stride, padding=padding, output_padding=output_padding, dilation=dilation, groups=groups)
  return mod(input, weight, bias)


def max_pool2d(input: Tensor, kernel_size, stride=None, padding=0, dilation=1,
               ceil_mode=False, return_indices=False):
  mod = modules.MaxPool2d(
    kernel_size=kernel_size, stride=stride, padding=padding, dilation=dilation,
    ceil_mode=ceil_mode, return_indices=return_indices)
  mod.as_returnn_torch_functional()
  return mod(input)


def relu(input: Tensor) -> Tensor:
  return modules.ReLU().as_returnn_torch_functional()(input)


def leaky_relu(input: Tensor, negative_slope: float = 0.01, inplace: bool = False) -> Tensor:
  return modules.LeakyReLU(negative_slope=negative_slope, inplace=inplace).as_returnn_torch_functional()(input)


def tanh(input: Tensor) -> Tensor:
  return modules.Tanh().as_returnn_torch_functional()(input)


def softmax(input: Tensor, dim: Optional[int] = None, dtype=None):
  return modules.Softmax(dim=dim).as_returnn_torch_functional()(input)


def abs(input: Tensor) -> Tensor:
  return modules.Abs().as_returnn_torch_functional()(input)


def log_softmax(input: Tensor, dim: Optional[int] = None, dtype=None):
  return modules.LogSoftmax(dim=dim).as_returnn_torch_functional()(input)


def log(input: Tensor):
  return modules.Log().as_returnn_torch_functional()(input)


def sigmoid(input: Tensor):
  return modules.Sigmoid().as_returnn_torch_functional()(input)


def logsigmoid(input: Tensor):
  return modules.LogSigmoid().as_returnn_torch_functional()(input)


def normalize(input: Tensor, p=2, dim=1, eps=1e-12) -> Tensor:
  norm_ = modules.Norm(p=p, axes=[dim], keepdims=True)(input)
  norm_f = modules.Reciprocal(eps=eps)(norm_)
  return input * norm_f


def norm(input: Tensor,
         p: Optional[Union[str, float, int]] = "fro",
         dim: Optional[Union[int, List[int]]] = None,
         keepdim: bool = False) -> Tensor:
  return modules.Norm(p=p, axes=[dim], keepdims=keepdim)(input)


def norm_except_dim(v: Tensor, pow: int = 2, dim: int = 0) -> Tensor:
  return modules.Norm(p=pow, axes=[i for i in range(v.dim()) if i != dim], keepdims=True)(v)




import _setup_test_env  # noqa
import sys
import unittest
import typing
import numpy
import tensorflow as tf
from pytorch_to_returnn import torch
from pytorch_to_returnn.converter import verify_torch_and_convert_to_returnn
from pytorch_to_returnn.pprint import pformat


def test_embedding():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    model = torch.nn.Embedding(n_in, n_out)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.randint(0, n_in, (n_time, n_batch), dtype="int64")
  verify_torch_and_convert_to_returnn(
    model_func,
    inputs=x, inputs_data_kwargs={"shape": (None,), "sparse": True, "dim": n_in, "batch_dim_axis": 1})


def test_linear_multiple_steps():
  n_steps = 3
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    ins = inputs.chunk(n_steps, dim=-1)
    model = torch.nn.Linear(n_in, n_out)
    outs = [model(x) for x in ins]
    out = sum(outs)
    return out

  x = numpy.ones((n_batch, n_time, n_in * n_steps)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x, inputs_data_kwargs={
      "shape": (None, n_in * n_steps), "batch_dim_axis": 0, "time_dim_axis": 1, "feature_dim_axis": 2})


def test_conv():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    # {'class': 'transposed_conv', 'from': 'layer2', 'activation': None, 'with_bias': True,
    #  'n_out': 192, 'filter_size': (10,), 'strides': (5,), 'remove_padding': (3,), 'output_padding': (1,)}
    model = torch.nn.Conv1d(
      in_channels=n_in,
      out_channels=n_out,
      kernel_size=3,
      stride=2)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_conv2d():
  n_in, n_out = 11, 13
  n_batch, n_time1, n_time2 = 3, 17, 19

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    # {'class': 'transposed_conv', 'from': 'layer2', 'activation': None, 'with_bias': True,
    #  'n_out': 192, 'filter_size': (10,), 'strides': (5,), 'remove_padding': (3,), 'output_padding': (1,)}
    model = torch.nn.Conv2d(
      in_channels=n_in,
      out_channels=n_out,
      kernel_size=(3, 5),
      stride=2)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time1, n_time2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_conv_transposed():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    # {'class': 'transposed_conv', 'from': 'layer2', 'activation': None, 'with_bias': True,
    #  'n_out': 192, 'filter_size': (10,), 'strides': (5,), 'remove_padding': (3,), 'output_padding': (1,)}
    model = torch.nn.ConvTranspose1d(
      in_channels=n_in,
      out_channels=n_out,
      kernel_size=10,
      stride=5,
      padding=3,
      output_padding=1)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_conv_transposed_2d():
  n_in, n_out = 11, 13
  n_batch, n_time1, n_time2 = 3, 17, 19

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    model = torch.nn.ConvTranspose2d(
      in_channels=n_in,
      out_channels=n_out,
      kernel_size=(10, 3),
      stride=5,
      padding=(2, 3),
      output_padding=1)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time1, n_time2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_functional_conv():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7
  kernel_size = 3

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_out, n_in, kernel_size)).astype("float32")
    bias = rnd.normal(0., 1., (n_out,)).astype("float32")
    weight = torch.from_numpy(weight)
    bias = torch.from_numpy(bias)
    return F.conv1d(
      inputs,
      weight=weight, bias=bias,
      stride=2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_functional_conv_no_bias():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7
  kernel_size = 3

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_out, n_in, kernel_size)).astype("float32")
    weight = torch.from_numpy(weight)
    return F.conv1d(inputs, weight=weight, stride=2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_functional_conv2d():
  n_in, n_out = 11, 13
  n_batch, n_time1, n_time2 = 3, 17, 19
  kernel_size = (3, 5)

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_out, n_in) + kernel_size).astype("float32")
    bias = rnd.normal(0., 1., (n_out,)).astype("float32")
    weight = torch.from_numpy(weight)
    bias = torch.from_numpy(bias)
    return F.conv2d(
      inputs,
      weight=weight, bias=bias,
      stride=2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time1, n_time2)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_functional_conv_transposed():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7
  kernel_size = 3

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")
    rnd = numpy.random.RandomState(42)
    weight = rnd.normal(0., 1., (n_in, n_out, kernel_size)).astype("float32")
    weight = torch.from_numpy(weight)
    return F.conv_transpose1d(inputs, weight=weight, stride=2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_batch_norm():
  n_in, n_batch, n_time = 11, 3, 7
  for train in [True, False]:

    def model_func(wrapped_import, inputs: torch.Tensor):
      if typing.TYPE_CHECKING or not wrapped_import:
        import torch
      else:
        torch = wrapped_import("torch")
      model = torch.nn.BatchNorm1d(n_in)
      if not train:
        model.eval()
      out = model(inputs)
      if train:
        model.reset_running_stats()  # for the test, such that we start with initial running mean/var
      return out

    x = numpy.ones((n_batch, n_in, n_time)).astype("float32")
    verify_torch_and_convert_to_returnn(model_func, inputs=x, train=train)

    rnd = numpy.random.RandomState(42)
    x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
    verify_torch_and_convert_to_returnn(model_func, inputs=x, train=train)


def test_batch_norm_running_stats():
  from pytorch_to_returnn.torch.nn import Module as ModuleReturnn
  from pytorch_to_returnn.naming import Naming, ModuleEntry
  from torch.nn import Module as ModuleTorch
  n_in, n_batch, n_time = 11, 3, 7
  mean_torch = None
  mean_returnn = None

  def model_func(wrapped_import, inputs: torch.Tensor):
    nonlocal mean_torch, mean_returnn
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    model = torch.nn.BatchNorm1d(n_in)
    out = model(inputs)
    if isinstance(model, ModuleTorch):
      mean_torch = model.running_mean.detach().cpu().numpy().copy()
    elif isinstance(model, ModuleReturnn):
      naming = Naming.get_instance()
      if naming.import_params_from_torch_namespace:  # only then we have the params
        module_entry = naming.modules[model]
        assert isinstance(module_entry, ModuleEntry)
        assert len(module_entry.calls) == 1
        call = module_entry.calls[0]
        assert call.returnn_layer
        mean_returnn = tf.squeeze(call.returnn_layer.params["batch_norm/mean"]).eval()
    model.reset_running_stats()  # for the test, such that we start with initial running mean/var
    return out

  x = numpy.ones((n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, train=True)
  assert mean_returnn is not None and mean_torch is not None
  print(mean_torch)
  numpy.testing.assert_allclose(mean_torch[0], 0.1, rtol=0, atol=1e-5)  # default momentum 0.1
  numpy.testing.assert_allclose(mean_returnn, mean_torch, rtol=0, atol=1e-5)

  mean_returnn = mean_torch = None
  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x, train=True)
  assert mean_returnn is not None and mean_torch is not None
  print(mean_torch)
  numpy.testing.assert_allclose(mean_returnn, mean_torch, rtol=0, atol=1e-5)


def test_unsqueeze():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    return inputs.unsqueeze(2)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_unsqueeze2():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    return inputs.unsqueeze(3)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_movedim():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    x = inputs  # (B,F,T)
    x = torch.nn.Conv1d(n_in, n_out, 3)(x)  # make it (B,T,F) in RETURNN. (B,F,T) in Torch.
    x = torch.movedim(x, 0, 1)  # stay (B,T,F) in RETURNN. (F,B,T) in Torch.
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_flatten_batch():
  n_in, n_out = 11, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    x = inputs  # (B,F,T)
    x = torch.movedim(x, 1, 0)  # (F,B,T)
    x = torch.reshape(x, (x.shape[0], -1))  # (F,B*T)
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  converter = verify_torch_and_convert_to_returnn(model_func, inputs=x)
  assert converter.returnn_net_dict["Flatten"]["class"] == "flatten_batch", pformat(converter.returnn_net_dict)


def test_slice_1d():
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    x = inputs[0]  # (T)
    x = x[:3]
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_slice_2d():
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    x = inputs[:, 3:]
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_expand():
  n_batch, n_feat = 5, 1

  def model_func(wrapped_import, inputs: torch.Tensor):
    x = inputs.expand(3, 2, -1, 7)
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_feat)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_const_with_batch_and_gather():
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    batch_size = inputs.shape[0]
    x = inputs.new_zeros((2, batch_size, 13))  # const, shape (2,B,13). in RETURNN (B,2,13)
    x = x[0]  # shape (B,13). in RETURNN (B,13).
    x = x + inputs[:, :1]  # (B,13). just do this to have dep on inputs
    return x

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_lstm_2l():
  n_batch, n_time = 3, 7
  n_in, n_out = 11, 13
  n_layers = 2

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")
    batch_size = inputs.shape[1]
    hidden = inputs.new_zeros((n_layers, batch_size, n_out))  # const, shape (2,B,13). in RETURNN (B,2,13)
    hidden = (hidden, hidden)
    lstm = torch.nn.LSTM(n_in, n_out, n_layers)
    out, hidden_ = lstm(inputs, hidden)
    return out

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_in)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x,
    inputs_data_kwargs={"shape": (None, n_in), "batch_dim_axis": 1})


def test_multiple_outputs():
  n_batch, n_time = 3, 7
  n_in, n_out = 11, 13
  n_layers = 1

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    class InnerModel(torch.nn.Module):
      def __init__(self):
        super(InnerModel, self).__init__()
        self.lstm = torch.nn.LSTM(n_in, n_out, n_layers)

      def forward(self, x, hidden):
        return self.lstm(x, hidden)

    # Wrap InnerModel in another model, to make sure InnerModel is not on the flattened root namespace.
    class OuterModel(torch.nn.Module):
      def __init__(self):
        super(OuterModel, self).__init__()
        self.model = InnerModel()

      def forward(self, input):
        batch_size = input.shape[1]
        hidden = torch.zeros((n_layers, batch_size, n_out))  # const, shape (1,B,13). in RETURNN (B,1,13)
        hidden = (hidden, hidden)
        out, (h, c) = self.model(input, hidden)  # out is (T,B,out), h/c is (1,B,out)
        return out[0] + h[0] + c[0]  # combine all, to set dependencies. (B,out)

    model = OuterModel()
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_time, n_batch, n_in)).astype("float32")
  verify_torch_and_convert_to_returnn(
    model_func, inputs=x,
    inputs_data_kwargs={"shape": (None, n_in), "batch_dim_axis": 1})


def test_gated_linear_unit_via_chunk():
  n_in, n_out = 12, 13
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
    else:
      torch = wrapped_import("torch")

    class GLU(torch.nn.Module):
      def __init__(self, dim):
        super().__init__()
        self.dim = dim

      def forward(self, x):
        out, gate = x.chunk(2, dim=self.dim)
        return out * gate.sigmoid()

    model = GLU(dim=1)
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


def test_depth_wise_conv1d():
  n_in, n_out = 12, 24
  n_batch, n_time = 3, 7

  def model_func(wrapped_import, inputs: torch.Tensor):
    if typing.TYPE_CHECKING or not wrapped_import:
      import torch
      import torch.nn.functional as F
    else:
      torch = wrapped_import("torch")
      F = wrapped_import("torch.nn.functional")

    class DepthWiseConv1d(torch.nn.Module):
      def __init__(self, chan_in, chan_out, kernel_size, padding):
        super().__init__()
        self.padding = padding
        self.conv = torch.nn.Conv1d(chan_in, chan_out, kernel_size, groups=chan_in)

      def forward(self, x):
        x = F.pad(x, self.padding)
        return self.conv(x)

    model = DepthWiseConv1d(n_in, n_out, [3], (0, 0))
    return model(inputs)

  rnd = numpy.random.RandomState(42)
  x = rnd.normal(0., 1., (n_batch, n_in, n_time)).astype("float32")
  verify_torch_and_convert_to_returnn(model_func, inputs=x)


if __name__ == "__main__":
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




from __future__ import annotations
import math
from typing import Dict, Any
import tensorflow as tf
from returnn.tf.layers.basic import LinearLayer
from .module import Module
from ..parameter import Parameter
from ...tensor import Tensor
from .. import init


class Identity(Module):
  def __init__(self, *args, **kwargs):
    super(Identity, self).__init__()

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {"class": "copy", "from": self._get_input_layer_name(input)}


class Linear(Module):
  in_features: int
  out_features: int
  weight: Tensor

  def __init__(self, in_features: int, out_features: int, bias: bool = True) -> None:
    super(Linear, self).__init__()
    self.in_features = in_features
    self.out_features = out_features
    self.weight = Parameter(Tensor(out_features, in_features))
    if bias:
      self.bias = Parameter(Tensor(out_features))
    else:
      self.register_parameter('bias', None)
    self.reset_parameters()

  def reset_parameters(self) -> None:
    init.kaiming_uniform_(self.weight, a=math.sqrt(5))
    if self.bias is not None:
      fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
      bound = 1 / math.sqrt(fan_in)
      init.uniform_(self.bias, -bound, bound)

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "linear", "from": self._get_input_layer_name(input),
      "n_out": self.out_features,
      "with_bias": self.bias is not None,
      "activation": None}

  def check_returnn_layer(self, layer: LinearLayer):
    assert layer.input_data.dim == self.in_features

  def import_params_torch_to_returnn(self, *, layer: LinearLayer, torch_module: Linear):
    session = tf.compat.v1.get_default_session()
    values = torch_module.weight.detach().numpy()
    values = values.transpose()
    layer.params["W"].load(values, session=session)
    if self.bias is not None:
      layer.params["b"].load(torch_module.bias.detach().numpy(), session=session)


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




import _setup_test_env  # noqa
import sys
import unittest
import typing
import numpy
import torch
from pytorch_to_returnn import torch as torch_returnn
from pytorch_to_returnn.import_wrapper import wrapped_import_torch_traced


def assert_equal(msg, a, b):
  print(f"check {msg} {a!r} == {b!r}")
  assert a == b


def test_torch_traced_wrapped_tensor_new():
  torch_traced = wrapped_import_torch_traced("torch")
  x = torch.Tensor()
  x_ = torch_traced.Tensor()
  assert_equal("new_zeros(3).shape:", x.new_zeros(3).shape, x_.new_zeros(3).shape)
  assert_equal("new_zeros([3]).shape:", x.new_zeros([3]).shape, x_.new_zeros([3]).shape)
  assert_equal("new(3).shape:", x.new(3).shape, x_.new(3).shape)


if __name__ == "__main__":
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




import _setup_test_env  # noqa
import sys
import unittest
import typing
import numpy
from returnn.tf.network import TFNetwork, ExternData
from returnn.tf.layers.basic import InternalLayer
from returnn.tf.util.data import Data
from pytorch_to_returnn import torch
from pytorch_to_returnn.naming import Naming


def test_base_get_output_shape_from_returnn_conv2d_static():
  with Naming.make_instance() as naming:
    assert isinstance(naming, Naming)
    x = torch.Tensor(64, 1, 11, 13)
    x_ = naming.register_tensor(x)
    x_.returnn_data = Data(name="x", shape=(1, 11, 13), feature_dim_axis=1)
    x_.returnn_axis_from_torch_axis = {0: 0, 1: 1, 2: 2, 3: 3}

    net = TFNetwork(extern_data=ExternData())
    # E.g. conv layer, with padding "valid", kernel size 3.
    layer = InternalLayer(name="layer", network=net, out_type={"shape": (9, 11, 32)})

    torch_shape, returnn_axis_from_torch_axis = torch.nn.Module._base_get_output_shape_from_returnn(
      inputs_flat=[x], layer=layer)
    assert returnn_axis_from_torch_axis == {0: 0, 1: 3, 2: 1, 3: 2}
    assert torch_shape == (64, 32, 9, 11)


def test_base_get_output_shape_from_returnn_conv2d_dynamic():
  with Naming.make_instance() as naming:
    assert isinstance(naming, Naming)
    x = torch.Tensor(64, 1, 11, 13)
    x_ = naming.register_tensor(x)
    x_.returnn_data = Data(name="x", shape=(1, None, None), feature_dim_axis=1)
    x_.returnn_axis_from_torch_axis = {0: 0, 1: 1, 2: 2, 3: 3}

    net = TFNetwork(extern_data=ExternData())
    # E.g. conv layer, with padding "same".
    layer = InternalLayer(name="layer", network=net, out_type={"shape": (None, None, 32)})

    torch_shape, returnn_axis_from_torch_axis = torch.nn.Module._base_get_output_shape_from_returnn(
      inputs_flat=[x], layer=layer)
    assert returnn_axis_from_torch_axis == {0: 0, 1: 3, 2: 1, 3: 2}
    assert torch_shape == (64, 32, 11, 13)


def test_base_get_output_shape_from_returnn_2d_reorder_dynamic():
  with Naming.make_instance() as naming:
    assert isinstance(naming, Naming)
    x = torch.Tensor(64, 1, 11, 13)
    x_ = naming.register_tensor(x)
    x_.returnn_data = Data(name="x", shape=(1, None, None), feature_dim_axis=1, auto_create_placeholders=True)
    x_.returnn_axis_from_torch_axis = {0: 0, 1: 1, 2: 2, 3: 3}
    y_data = x_.returnn_data.copy_move_axis(2, 3)
    assert y_data.get_dim_tag(3) == x_.returnn_data.get_dim_tag(2)

    net = TFNetwork(extern_data=ExternData())
    # E.g. softmax_over_spatial with axis="stag:time1"
    layer = InternalLayer(name="layer", network=net, output=y_data)

    # We expect from all Torch modules, that they don't reorder the spatial axes.
    # (If they do, they explicitly would overwrite the output shape logic.)
    torch_shape, returnn_axis_from_torch_axis = torch.nn.Module._base_get_output_shape_from_returnn(
      inputs_flat=[x], layer=layer)
    assert returnn_axis_from_torch_axis == {0: 0, 1: 1, 2: 3, 3: 2}
    assert torch_shape == (64, 1, 11, 13)


if __name__ == "__main__":
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




from collections import OrderedDict, Counter
import types
from typing import Tuple
import importlib
import sys
from .context import WrapCtx
from .base_wrappers import WrappedObject, WrappedModule, WrappedSourceModule
from .base_wrappers import make_wrapped_object, make_wrapped_class, make_wrapped_function


def wrap(obj, *, name: str, ctx: WrapCtx):
  if isinstance(obj, (WrappedObject, WrappedModule)):
    return obj
  if isinstance(obj, ctx.keep_as_is_types):
    return obj
  obj = _nested_transform(obj, lambda _x: wrap(_x, name="%s..." % name, ctx=ctx))

  if isinstance(obj, types.ModuleType):
    if ctx.should_wrap_mod(obj.__name__):
      if obj.__name__ not in sys.modules:
        # If this happens, this is actually a bug in how the module importing is done.
        # This is not on our side.
        # This might happen for native Torch modules.
        # This is slightly problematic for our following logic, because import_module will not find it.
        # So we patch this here.
        sys.modules[obj.__name__] = obj
      # If this comes from an WrappedIndirectModule, this will create a new WrappedIndirectModule for the sub mod.
      # See logic in _MetaPathLoader.
      obj = importlib.import_module(ctx.wrapped_mod_prefix + obj.__name__)
  elif isinstance(obj, _FuncTypes):
    if not obj.__module__ or ctx.should_wrap_mod(obj.__module__):
      obj = make_wrapped_function(func=obj, name=name, ctx=ctx)
  elif isinstance(obj, type):
    if obj in ctx.explicit_wrapped_types:
      obj = ctx.explicit_wrapped_types[obj].new_type
    elif obj in ctx.keep_as_is_types:
      pass  # keep as-is
    elif type(obj) != type:  # explicitly do not check sub-types, e.g. pybind11_type
      pass
    elif obj.__module__ and ctx.should_wrap_mod(obj.__module__):
      # If this is an indirect module, but the type is part of a submodule which we want to transform,
      # explicitly check the import.
      mod_ = importlib.import_module(ctx.wrapped_mod_prefix + obj.__module__)
      if isinstance(mod_, WrappedSourceModule):
        obj = getattr(mod_, obj.__qualname__)
      else:
        obj = make_wrapped_class(cls=obj, name=name, ctx=ctx)
  elif not isinstance(obj, type) and type(type(obj)) == type:  # object instance
    if type(obj) in ctx.explicit_wrapped_types:
      obj = make_wrapped_object(obj, name=name, ctx=ctx)
    elif not getattr(type(obj), "__module__", None) or not ctx.should_wrap_mod(type(obj).__module__):
      pass  # Don't wrap this object.
    elif isinstance(obj, ctx.keep_as_is_types):  # blacklist
      pass  # keep as-is
    else:
      obj = make_wrapped_object(obj, name=name, ctx=ctx)

  return obj


def unwrap(obj):
  """
  :param obj: potentially wrapped object. could be nested.
  :return: unwrapped object.
    Note that this is not exactly the inverse of :func:`wrap`.
    Namely, not all wrapped objects are unwrapped,
    as this is also not possible in all cases.
  """
  obj = _nested_transform(obj, unwrap)
  if isinstance(obj, WrappedObject):
    # noinspection PyProtectedMember
    return obj._wrapped__orig_obj
  return obj  # leave as-is


def _nested_transform(obj, transform):
  if type(obj) == tuple:
    return tuple([transform(item) for item in obj])
  assert not isinstance(obj, tuple)  # namedtuple or so not implemented yet...
  if type(obj) == list:
    return [transform(item) for item in obj]
  assert not isinstance(obj, list)  # custom subclasses of list not implemented yet...
  if type(obj) == dict:
    # Assume keys are not wrapped.
    return {key: transform(value) for (key, value) in obj.items()}
  if type(obj) == dict:
    # Assume keys are not wrapped.
    return {key: transform(value) for (key, value) in obj.items()}
  if type(obj) == OrderedDict:
    return OrderedDict([(key, transform(value)) for (key, value) in obj.items()])
  if type(obj) == Counter:
    return obj  # as-is
  assert not isinstance(obj, dict)
  return obj


# https://youtrack.jetbrains.com/issue/PY-45382
# noinspection PyTypeChecker
_FuncTypes = (types.FunctionType, types.BuiltinFunctionType)  # type: Tuple[type]




"""
Only provides types used for type checking, and auto-completion, etc.

Note that this `naming` submodule can be used for different `torch` variants
(e.g. traced original, or our Torch-RETURNN wrapper).
"""

import typing
from typing import Union


if typing.TYPE_CHECKING:
  import torch
  from pytorch_to_returnn.torch.tensor import Tensor as TorchReturnnTensor
  from pytorch_to_returnn.torch.nn import Module as TorchReturnnModule

  # These are the two supported/common cases currently, but this might be extended.
  Tensor = Union[TorchReturnnTensor, torch.Tensor]
  Module = Union[TorchReturnnModule, torch.nn.Module]


# We don't provide the types if not type-checking, by intention.
# This should never be used in actual code (e.g. isinstance checks or so).
# You should always explicitly use the right type in such cases.




from typing import Any
import weakref
from collections import OrderedDict


class RemovableHandle(object):
  """A handle which provides the capability to remove a hook."""

  id: int
  next_id: int = 0

  def __init__(self, hooks_dict: Any) -> None:
    self.hooks_dict_ref = weakref.ref(hooks_dict)
    self.id = RemovableHandle.next_id
    RemovableHandle.next_id += 1

  def remove(self) -> None:
    hooks_dict = self.hooks_dict_ref()
    if hooks_dict is not None and self.id in hooks_dict:
      del hooks_dict[self.id]

  def __getstate__(self):
    return (self.hooks_dict_ref(), self.id)

  def __setstate__(self, state) -> None:
    if state[0] is None:
      # create a dead reference
      self.hooks_dict_ref = weakref.ref(OrderedDict())
    else:
      self.hooks_dict_ref = weakref.ref(state[0])
    self.id = state[1]
    RemovableHandle.next_id = max(RemovableHandle.next_id, self.id + 1)

  def __enter__(self) -> 'RemovableHandle':
    return self

  def __exit__(self, type: Any, value: Any, tb: Any) -> None:
    self.remove()




import sys
import types
from typing import Union, Any
import importlib
from .base_wrappers import WrappedModule
from .context import WrapCtx
from .meta_path import MetaPathFinder, MetaPathLoader
from .. import log


WrappedModPrefixes = set()


def import_module(mod_name: str, *, ctx: WrapCtx) -> Union[types.ModuleType, WrappedModule, Any]:
  _maybe_register_meta_path(ctx)
  if ctx.mod_map.should_wrap_mod_name(mod_name):
    mapped_mod_name = ctx.mod_map.map_mod_name(mod_name)
  else:
    mapped_mod_name = ctx.wrapped_mod_prefix + mod_name
  mod = importlib.import_module(mapped_mod_name)
  if mapped_mod_name.startswith(ctx.wrapped_mod_prefix):
    assert isinstance(mod, WrappedModule)
  assert isinstance(mod, types.ModuleType)
  return mod


def _maybe_register_meta_path(ctx: WrapCtx):
  for meta_path in sys.meta_path:
    if isinstance(meta_path, MetaPathFinder):
      if meta_path.ctx is ctx:  # TODO?
        return

  if log.Verbosity >= 4:
    print(f"*** register sys.meta_path for ctx {ctx}")
  loader = MetaPathLoader(ctx=ctx)
  finder = MetaPathFinder(loader=loader)
  sys.meta_path.append(finder)
  WrappedModPrefixes.add(ctx.wrapped_mod_prefix)




from typing import Dict, Any, Optional
from returnn.tf.layers.basic import PoolLayer
from .module import Module
from .utils import _ntuple
from ..common_types import _size_any_t
from ...tensor import Tensor


class _MaxPoolNd(Module):
  nd: Optional[int] = None  # set in subclass
  return_indices: bool
  ceil_mode: bool

  def __init__(self, kernel_size: _size_any_t, stride: Optional[_size_any_t] = None,
               padding: _size_any_t = 0, dilation: _size_any_t = 1,
               return_indices: bool = False, ceil_mode: bool = False) -> None:
    super(_MaxPoolNd, self).__init__()
    self.kernel_size = _ntuple(self.nd)(kernel_size)
    self.stride = _ntuple(self.nd)(stride) if (stride is not None) else _ntuple(self.nd)(kernel_size)
    self.padding = _ntuple(self.nd)(padding)
    self.dilation = _ntuple(self.nd)(dilation)
    self.return_indices = return_indices
    self.ceil_mode = ceil_mode

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    assert not self.return_indices  # not implemented
    assert not self.ceil_mode  # not implemented (or maybe ignore?)
    assert all(p == 0 for p in self.padding)  # not implemented
    return {
      "class": "pool", "mode": "max", "from": self._get_input_layer_name(input),
      "pool_size": self.kernel_size,
      "dilation_rate": self.dilation,
      "strides": self.stride,
      "padding": "valid"}


class MaxPool1d(_MaxPoolNd):
  nd = 1


class MaxPool2d(_MaxPoolNd):
  nd = 2


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




"""
This file is originally called "flatten.py", but we extend it by further reshape modules.

Note on PyTorch vs RETURNN/TF terminology:

*Axis* in TF == *dim* in PyTorch.
*Dim(s)* in TF == *size* in PyTorch.
"""

from __future__ import annotations
from typing import Tuple, List, Union, Collection, Dict, Any, Optional
from functools import reduce
import operator
from returnn.tf.layers.basic import LayerBase, MergeDimsLayer, FlattenBatchLayer
from .module import Module
from ...tensor import Tensor, Size
from ....naming import Naming


_NamedShape = Tuple[Tuple[str, int]]


class MergeDims(Module):
  """
  Maps to RETURNN MergeDimsLayer, if batch axis is not involved.
  In case batch axis is involved, this uses RETURNN FlattenBatchLayer.

  Used by Flatten, and thus indirectly by various kinds of reshape.
  """
  is_original_torch_module = False

  def __init__(self, dims: Collection[int]):
    super(MergeDims, self).__init__()
    self.dims = dims

  @classmethod
  def generic_create_returnn_layer_dict(cls, input: Tensor, dims: Collection[int]) -> Dict[str, Any]:
    naming = Naming.get_instance()
    in_ = naming.register_tensor(input)
    assert in_.returnn_data
    assert isinstance(dims, (tuple, list))

    def _pos_dim(d: int) -> int:
      assert -len(input.shape) <= d < len(input.shape)
      if d < 0:
        d += len(input.shape)
      assert 0 <= d < len(input.shape)
      return d

    dims = [_pos_dim(d) for d in dims]
    returnn_axes = [in_.returnn_axis_from_torch_axis[d] for d in dims]
    if in_.returnn_data.have_batch_axis() and in_.returnn_data.batch_dim_axis in returnn_axes:
      if set(in_.returnn_data.get_dynamic_axes()).intersection(returnn_axes):
        # We merge batch with some other dynamic axis.
        batch_major = returnn_axes[0] == in_.returnn_data.batch_dim_axis
        returnn_axes.pop(in_.returnn_data.batch_dim_axis)
        assert len(returnn_axes) == 1  # not implemented otherwise
        return {
          "class": "flatten_batch",
          "from": cls._get_input_layer_name(input),
          "axis": cls._get_input_axis_to_returnn(input, axis=in_.torch_axis_from_returnn_axis[returnn_axes[0]]),
          "batch_major": batch_major}

    return {
      "class": "merge_dims", "from": cls._get_input_layer_name(input),
      "axes": [cls._get_input_axis_to_returnn(input, axis=axis) for axis in dims],
      "keep_order": True}

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return self.generic_create_returnn_layer_dict(input=input, dims=self.dims)

  @classmethod
  def generic_get_output_shape_from_returnn(cls,
                                            input: Tensor,
                                            layer: Union[MergeDimsLayer, FlattenBatchLayer],
                                            dims: Collection[int]
                                            ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    We can exactly infer this, so we don't need to rely on the default heuristics.
    Also, we must do this, in case of dynamic axes.

    :return: (torch_shape, returnn_axis_from_torch_axis).
    """
    naming = Naming.get_instance()
    in_ = naming.register_tensor(input)
    assert in_.returnn_data
    assert isinstance(dims, (tuple, list))

    def _pos_dim(d: int) -> int:
      assert -len(input.shape) <= d < len(input.shape)
      if d < 0:
        d += len(input.shape)
      assert 0 <= d < len(input.shape)
      return d

    # Switch to TF terminology. axis, not dim.
    in_torch_axes = [_pos_dim(d) for d in dims]
    in_returnn_axes = [in_.returnn_axis_from_torch_axis[i] for i in in_torch_axes]
    out_returnn_reduced_axis = min(in_returnn_axes)
    out_size = reduce(operator.mul, [input.shape[i] for i in in_torch_axes], 1)
    out_torch_reduced_axis = min(in_torch_axes)
    out_torch_shape = [input.shape[i] for i in range(input.ndim) if i not in in_torch_axes]
    out_torch_shape = tuple(
      out_torch_shape[:out_torch_reduced_axis] + [out_size] + out_torch_shape[out_torch_reduced_axis:])

    out_returnn_axis_from_torch_axis = {out_torch_reduced_axis: out_returnn_reduced_axis}
    for in_torch_axis in range(len(input.shape)):
      if in_torch_axis in in_torch_axes:
        continue  # reduced. covered already
      elif in_torch_axis < min(in_torch_axes):
        torch_offset = 0
      elif in_torch_axis > max(in_torch_axes):
        torch_offset = len(in_torch_axes) - 1
      else:
        torch_offset = len([a for a in in_torch_axes if a < in_torch_axis])
        assert 0 < torch_offset < len(in_torch_axes), f"in_torch_axis={in_torch_axis}?"  # min/max already covered
        torch_offset -= 1
      out_torch_axis = in_torch_axis - torch_offset
      assert out_torch_axis not in out_returnn_axis_from_torch_axis

      in_returnn_axis = in_.returnn_axis_from_torch_axis[in_torch_axis]
      assert in_returnn_axis not in in_returnn_axes  # not reduced axis
      if in_returnn_axis < min(in_returnn_axes):
        returnn_offset = 0
      elif in_returnn_axis > max(in_returnn_axes):
        returnn_offset = len(in_returnn_axes) - 1
      else:
        returnn_offset = len([a for a in in_returnn_axes if a < in_returnn_axis])
        assert 0 < returnn_offset < len(in_returnn_axes), f"in_returnn_axis={in_returnn_axis}?"
        returnn_offset -= 1
      out_returnn_axis = in_returnn_axis - returnn_offset
      out_returnn_axis_from_torch_axis[out_torch_axis] = out_returnn_axis

    if isinstance(layer, MergeDimsLayer):
      pass
    elif isinstance(layer, FlattenBatchLayer):
      # The reduced axis was moved to front, i.e. out_returnn_reduced_axis = 0.
      if out_returnn_reduced_axis != 0:  # only needs change if it was not 0 already
        # Basically movedim(out, out_returnn_axis_from_torch_axis, 0).
        def _new_axis(old: int) -> int:
          if old == out_returnn_reduced_axis:
            return 0
          if old < out_returnn_reduced_axis:
            return old + 1
          return old
        out_returnn_axis_from_torch_axis = {i: _new_axis(j) for (i, j) in out_returnn_axis_from_torch_axis.items()}
    else:
      raise TypeError(f"Unexpected layer {layer}, with input {input} and dims {dims}")

    return out_torch_shape, out_returnn_axis_from_torch_axis

  def _get_output_shape_from_returnn(self,
                                     inputs_flat: List[Tensor], layer: Union[MergeDimsLayer, FlattenBatchLayer]
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    assert len(inputs_flat) == 1
    return self.generic_get_output_shape_from_returnn(input=inputs_flat[0], layer=layer, dims=self.dims)


class Squeeze(Module):
  is_original_torch_module = False

  def __init__(self, dim: int):
    super(Squeeze, self).__init__()
    self.dim = dim

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    return {
      "class": "squeeze", "from": self._get_input_layer_name(input),
      "axis": self._get_input_axis_to_returnn(input, self.dim)}


class Flatten(Module):
  """
  Originally in flatten.py.
  """
  def __init__(self, start_dim: int = 1, end_dim: int = -1) -> None:
    super(Flatten, self).__init__()
    self.start_dim = start_dim
    self.end_dim = end_dim

  def _get_dims(self, input: Tensor):
    def _dim(d: int) -> int:
      assert -input.ndim <= d < input.ndim
      if d < 0:
        d += input.ndim
      assert 0 <= d < input.ndim
      return d
    start_dim = _dim(self.start_dim)
    end_dim = _dim(self.end_dim)
    assert start_dim <= end_dim
    dims = list(range(start_dim, end_dim + 1))  # end_dim is inclusive
    return dims

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    dims = self._get_dims(input)
    return MergeDims.generic_create_returnn_layer_dict(input=input, dims=dims)

  def _get_output_shape_from_returnn(self,
                                     inputs_flat: List[Tensor], layer: Union[MergeDimsLayer, FlattenBatchLayer]
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    assert len(inputs_flat) == 1
    input = inputs_flat[0]
    dims = self._get_dims(input)
    return MergeDims.generic_get_output_shape_from_returnn(input=input, layer=layer, dims=dims)


class Unflatten(Module):
  """
  Originally in flatten.py.
  """
  def __init__(self,
               dim: Union[int, str],
               unflattened_size: Union[Size, _NamedShape, Tuple[int, ...], List[int]]) -> None:
    super(Unflatten, self).__init__()
    self.dim = dim
    self.unflattened_size = unflattened_size

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    assert isinstance(self.dim, int)  # not implemented otherwise
    dims = list(self.unflattened_size)
    # Introduce -1 again to dims, such that we can handle dynamic axes in RETURNN.
    non_one_dims = [d for d in dims if d != 1]
    if len(non_one_dims) == 1:
      dims[dims.index(non_one_dims[0])] = -1
    return {
      "class": "split_dims", "from": self._get_input_layer_name(input),
      "axis": self._get_input_axis_to_returnn(input, self.dim),
      "dims": dims}

  def _get_output_shape_from_returnn(self,
                                     inputs_flat: List[Tensor], layer: LayerBase
                                     ) -> Tuple[Tuple[int, ...], Dict[int, int]]:
    """
    We can exactly infer this, so we don't need to rely on the default heuristics.

    :return: (torch_shape, returnn_axis_from_torch_axis).
    """
    naming = Naming.get_instance()
    assert len(inputs_flat) == 1
    x = inputs_flat[0]
    x_entry = naming.register_tensor(x)
    dim = self.dim
    assert -len(x.shape) <= dim < len(x.shape)
    if dim < 0:
      dim += len(x.shape)
    assert 0 <= dim < len(x.shape)
    unflatten_size = list(self.unflattened_size)
    if any(d == -1 for d in unflatten_size):
      rem_dim = x.shape[dim]
      for d in unflatten_size:
        if d == -1:
          continue
        assert rem_dim % d == 0
        rem_dim //= d
      unflatten_size[unflatten_size.index(-1)] = rem_dim
    assert all(d > 0 for d in unflatten_size)
    out_torch_shape = x.shape[:dim] + tuple(unflatten_size) + x.shape[dim + 1:]

    old_returnn_split_axis = x_entry.returnn_axis_from_torch_axis[dim]
    out_returnn_axis_from_torch_axis = {}
    for i in range(len(out_torch_shape)):
      unflatten_idx = None
      if i < dim:
        old_torch_axis = i
      elif i > dim + len(unflatten_size):
        old_torch_axis = i - len(unflatten_size) + 1
      else:  # i >= dim and ..., within the unflattened dims
        unflatten_idx = i - dim
        old_torch_axis = dim
      old_returnn_axis = x_entry.returnn_axis_from_torch_axis[old_torch_axis]
      if old_returnn_axis < old_returnn_split_axis:
        returnn_axis = old_returnn_axis
      elif old_returnn_axis > old_returnn_split_axis:
        returnn_axis = old_returnn_axis + len(unflatten_size) - 1
      else:
        returnn_axis = old_returnn_axis + unflatten_idx
      out_returnn_axis_from_torch_axis[i] = returnn_axis
    return out_torch_shape, out_returnn_axis_from_torch_axis


class SplitDims(Unflatten):
  """
  Alias, RETURNN consistent name
  """
  is_original_torch_module = False


class Split(Module):
  """
  tf.split, SplitLayer in RETURNN
  """
  is_original_torch_module = False

  def __init__(self, *, dim: int, num_splits: Optional[int] = None, size_splits: Optional[List[int]] = None):
    super(Split, self).__init__()
    self.dim = dim
    assert num_splits or size_splits
    self.num_splits = num_splits
    self.size_splits = size_splits

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    d = {
      "class": "split", "from": self._get_input_layer_name(input),
      "axis": self._get_input_axis_to_returnn(input, axis=self.dim)}
    if self.num_splits is not None:
      d["num_splits"] = self.num_splits
    if self.size_splits is not None:
      d["size_splits"] = self.size_splits
    return d

  def make_structured_returnn_output(self, output: Tensor) -> List[Tensor]:
    from .operator import GetSublayer
    if self.num_splits is not None:
      num_splits = self.num_splits
    else:
      num_splits = len(self.size_splits)
    return [GetSublayer(sub_layer=f"{i}")(output) for i in range(num_splits)]


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]




from __future__ import annotations
import math
import numbers
from typing import Dict, Any, Union, List
import tensorflow as tf
from returnn.tf.layers.basic import LayerNormLayer
from .module import Module
from ..parameter import Parameter
from ...tensor import Tensor, Size
from .. import init


_shape_t = Union[int, List[int], Size]


class LayerNorm(Module):
  def __init__(self, normalized_shape: _shape_t, eps: float = 1e-5, elementwise_affine: bool = True) -> None:
    super(LayerNorm, self).__init__()
    if isinstance(normalized_shape, numbers.Integral):
      normalized_shape = (normalized_shape,)
    self.normalized_shape = tuple(normalized_shape)
    self.eps = eps
    self.elementwise_affine = elementwise_affine
    if self.elementwise_affine:
      self.weight = Parameter(Tensor(*normalized_shape))
      self.bias = Parameter(Tensor(*normalized_shape))
    else:
      self.register_parameter('weight', None)
      self.register_parameter('bias', None)
    self.reset_parameters()

  def reset_parameters(self) -> None:
    if self.elementwise_affine:
      init.ones_(self.weight)
      init.zeros_(self.bias)

  def create_returnn_layer_dict(self, input: Tensor) -> Dict[str, Any]:
    assert len(self.normalized_shape) == 1  # not implemented otherwise
    axis = self._get_input_axis_to_returnn(input, axis=-1)
    assert axis == "F"  # not implemented otherwise. expect that feature-dim is last.
    assert self.elementwise_affine
    return {
      "class": "layer_norm", "from": self._get_input_layer_name(input),
      "epsilon": self.eps}

  def import_params_torch_to_returnn(self, *, layer: LayerNormLayer, torch_module: LayerNorm):
    assert isinstance(layer, LayerNormLayer)
    assert self.elementwise_affine
    assert len(self.normalized_shape) == 1  # not implemented otherwise
    session = tf.compat.v1.get_default_session()
    layer.params["scale"].load(torch_module.weight.detach().numpy(), session=session)
    layer.params["bias"].load(torch_module.bias.detach().numpy(), session=session)


__all__ = [
  key for (key, value) in sorted(globals().items())
  if not key.startswith("_")
  and getattr(value, "__module__", "") == __name__]
